import argparse
import json
import logging
from typing import Any

import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from birder.adversarial.deepfool import DeepFool
from birder.adversarial.fgsm import FGSM
from birder.adversarial.pgd import PGD
from birder.adversarial.simba import SimBA
from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.conf import settings
from birder.data.dataloader.webdataset import make_wds_loader
from birder.data.datasets.directory import make_image_dataset
from birder.data.datasets.webdataset import make_wds_dataset
from birder.data.datasets.webdataset import prepare_wds_args
from birder.data.datasets.webdataset import wds_args_from_info
from birder.data.transforms.classification import RGBType
from birder.data.transforms.classification import inference_preset

logger = logging.getLogger(__name__)


def _build_attack(
    method: str,
    net: torch.nn.Module,
    rgb_stats: RGBType,
    eps: float,
    steps: int,
    step_size: float | None,
    deepfool_num_classes: int,
) -> FGSM | PGD | DeepFool | SimBA:
    if method == "fgsm":
        return FGSM(net, eps=eps, rgb_stats=rgb_stats)

    if method == "pgd":
        return PGD(
            net,
            eps=eps,
            steps=steps,
            step_size=step_size,
            random_start=False,
            rgb_stats=rgb_stats,
        )

    if method == "deepfool":
        return DeepFool(net, num_classes=deepfool_num_classes, overshoot=0.02, max_iter=steps, rgb_stats=rgb_stats)

    if method == "simba":
        return SimBA(
            net,
            step_size=step_size if step_size is not None else eps,
            max_iter=steps,
            rgb_stats=rgb_stats,
        )

    raise ValueError(f"Unsupported attack method '{method}'")


# pylint: disable=too-many-locals,too-many-branches
def evaluate_robust(args: argparse.Namespace) -> None:
    if args.gpu is True:
        device = torch.device("cuda")
    elif args.mps is True:
        device = torch.device("mps")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    if args.amp_dtype is None:
        amp_dtype = torch.get_autocast_dtype(device.type)
        logger.debug(f"AMP: {args.amp}, AMP dtype: {amp_dtype}")
    else:
        amp_dtype = getattr(torch, args.amp_dtype)

    network_name = lib.get_network_name(args.network, tag=args.tag)
    net, model_info = fs_ops.load_model(
        device,
        args.network,
        tag=args.tag,
        epoch=args.epoch,
        new_size=args.size,
        inference=True,
        reparameterized=args.reparameterized,
    )

    class_to_idx = model_info.class_to_idx
    rgb_stats = model_info.rgb_stats
    if args.size is None:
        size = lib.get_size_from_signature(model_info.signature)
    else:
        size = args.size

    transform = inference_preset(size, rgb_stats, args.center_crop, args.simple_crop)

    if args.wds is True:
        wds_path: str | list[str]
        if args.wds_info is not None:
            wds_path, dataset_size = wds_args_from_info(args.wds_info, args.wds_split)
            if args.wds_size is not None:
                dataset_size = args.wds_size
        else:
            wds_path, dataset_size = prepare_wds_args(args.data_path[0], args.wds_size, device)

        num_samples = dataset_size
        dataset = make_wds_dataset(
            wds_path,
            dataset_size=dataset_size,
            shuffle=False,
            samples_names=True,
            transform=transform,
        )
        dataloader = make_wds_loader(
            dataset,
            args.batch_size,
            num_workers=args.num_workers,
            prefetch_factor=None,
            collate_fn=None,
            world_size=1,
            pin_memory=False,
            exact=True,
        )
    else:
        dataset = make_image_dataset(args.data_path, class_to_idx, transforms=transform)
        num_samples = len(dataset)
        dataloader = DataLoader(dataset, batch_size=args.batch_size, num_workers=args.num_workers)

    attack = _build_attack(args.method, net, rgb_stats, args.eps, args.steps, args.step_size, args.deepfool_num_classes)

    clean_correct = 0
    adv_correct = 0
    total = 0
    skipped_unlabeled = 0
    with tqdm(total=num_samples, unit="images", leave=False) as progress:
        for _, inputs, targets in dataloader:
            inputs = inputs.to(device)
            targets = targets.to(device)
            batch_size = inputs.size(0)

            valid_mask = targets != settings.NO_LABEL
            num_valid = int(valid_mask.sum().item())
            skipped_unlabeled += batch_size - num_valid
            if num_valid == 0:
                progress.update(batch_size)
                continue

            inputs = inputs[valid_mask]
            targets = targets[valid_mask]

            with torch.no_grad():
                with torch.amp.autocast(device.type, enabled=args.amp, dtype=amp_dtype):
                    clean_logits = net(inputs)

                clean_preds = clean_logits.argmax(dim=1)
                clean_correct += (clean_preds == targets).sum().item()

            result = attack(inputs, target=None)
            adv_logits = result.adv_logits
            adv_preds = adv_logits.argmax(dim=1)
            adv_correct += (adv_preds == targets).sum().item()

            total += num_valid
            progress.update(batch_size)

    if total == 0:
        raise RuntimeError(f"No labeled samples found (all labels are {settings.NO_LABEL})")

    if skipped_unlabeled > 0:
        logger.warning(f"Skipped {skipped_unlabeled} unlabeled samples (label={settings.NO_LABEL})")

    clean_accuracy = clean_correct / total
    adv_accuracy = adv_correct / total
    accuracy_drop = clean_accuracy - adv_accuracy

    logger.info(
        f"{network_name}: clean={clean_accuracy:.4f}, adv={adv_accuracy:.4f}, drop={accuracy_drop:.4f} "
        f"(evaluated on {total} labeled samples)"
    )

    if args.dry_run is False:
        output = {
            "method": "robust",
            "accuracy": adv_accuracy,
            "clean_accuracy": clean_accuracy,
            "accuracy_drop": accuracy_drop,
            "attack_method": args.method,
            "epsilon": args.eps,
            "num_samples": total,
            "num_skipped_unlabeled": skipped_unlabeled,
        }

        output_dir = settings.RESULTS_DIR.joinpath(args.dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        epoch_str = f"_e{args.epoch}" if args.epoch is not None else ""
        output_path = output_dir.joinpath(f"{network_name}{epoch_str}_{args.method}_eps{args.eps}.json")

        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(output, f, indent=2)

        logger.info(f"Results saved to {output_path}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "adversarial",
        allow_abbrev=False,
        help="evaluate adversarial robustness of a model on a dataset",
        description="evaluate adversarial robustness of a model on a dataset",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval adversarial -n resnet_v2_50 -e 100 --method pgd --eps 0.02 --gpu data/validation\n"
            "python -m birder.eval adversarial -n vovnet_v2_39 -t il-common --method pgd --batch-size 4 "
            "--gpu --gpu-id 1 --fast-matmul data/validation_il-common_packed\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("-n", "--network", type=str, help="neural network to evaluate")
    subparser.add_argument("-t", "--tag", type=str, help="model tag")
    subparser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint epoch")
    subparser.add_argument("--reparameterized", default=False, action="store_true", help="load reparameterized model")
    subparser.add_argument(
        "--method",
        type=str,
        choices=["fgsm", "pgd", "deepfool", "simba"],
        help="adversarial attack method",
    )
    subparser.add_argument("--eps", type=float, default=0.007, help="perturbation budget in pixel space [0, 1]")
    subparser.add_argument("--steps", type=int, default=10, help="number of iterations for iterative attacks")
    subparser.add_argument("--step-size", type=float, help="step size in pixel space (defaults to eps/steps for PGD)")
    subparser.add_argument(
        "--deepfool-num-classes", type=int, default=10, help="number of top classes to consider for DeepFool"
    )
    subparser.add_argument(
        "--size", type=int, nargs="+", metavar=("H", "W"), help="image size for inference (defaults to model signature)"
    )
    subparser.add_argument(
        "--amp", default=False, action="store_true", help="use torch.amp.autocast for mixed precision inference"
    )
    subparser.add_argument(
        "--amp-dtype",
        type=str,
        choices=["float16", "bfloat16"],
        help="whether to use float16 or bfloat16 for mixed precision",
    )
    subparser.add_argument(
        "--fast-matmul", default=False, action="store_true", help="use fast matrix multiplication (affects precision)"
    )
    subparser.add_argument("--batch-size", type=int, default=32, metavar="N", help="the batch size")
    subparser.add_argument(
        "-j", "--num-workers", type=int, default=8, metavar="N", help="number of preprocessing workers"
    )
    subparser.add_argument("--center-crop", type=float, default=1.0, help="center crop ratio to use during inference")
    subparser.add_argument(
        "--simple-crop",
        default=False,
        action="store_true",
        help="use a simple crop that preserves aspect ratio but may trim parts of the image",
    )
    subparser.add_argument(
        "--dir", type=str, default="robust", help="place all outputs in a sub-directory (relative to results)"
    )
    subparser.add_argument("--dry-run", default=False, action="store_true", help="skip saving results to file")
    subparser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    subparser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    subparser.add_argument(
        "--mps", default=False, action="store_true", help="use mps (Metal Performance Shaders) device"
    )
    subparser.add_argument("--wds", default=False, action="store_true", help="evaluate a webdataset directory")
    subparser.add_argument("--wds-size", type=int, metavar="N", help="size of the wds dataset")
    subparser.add_argument("--wds-info", type=str, metavar="FILE", help="wds info file path")
    subparser.add_argument(
        "--wds-split", type=str, default="validation", metavar="NAME", help="wds dataset split to load"
    )
    subparser.add_argument("data_path", nargs="*", help="data files path (directories and files)")
    subparser.set_defaults(func=main)


def validate_args(args: argparse.Namespace) -> None:
    args.size = cli.parse_size(args.size)
    if args.network is None:
        raise cli.ValidationError("--network is required")
    if args.method is None:
        raise cli.ValidationError("--method is required")
    if args.center_crop > 1 or args.center_crop <= 0.0:
        raise cli.ValidationError(f"--center-crop must be in range of (0, 1.0], got {args.center_crop}")

    if args.wds is False and len(args.data_path) == 0:
        raise cli.ValidationError("Must provide at least one data source, --data-path or --wds")

    if args.wds is True:
        if args.wds_info is None and len(args.data_path) == 0:
            raise cli.ValidationError("--wds requires a data path unless --wds-info is provided")
        if len(args.data_path) > 1:
            raise cli.ValidationError(f"--wds can have at most 1 --data-path, got {len(args.data_path)}")
        if args.wds_info is None and len(args.data_path) == 1:
            data_path = args.data_path[0]
            if "://" in data_path and args.wds_size is None:
                raise cli.ValidationError("--wds-size is required for remote --data-path")


def main(args: argparse.Namespace) -> None:
    validate_args(args)
    evaluate_robust(args)



"""
iBOT, adapted from
https://github.com/bytedance/ibot/blob/main/main_ibot.py

Paper "iBOT: Image BERT Pre-Training with Online Tokenizer", https://arxiv.org/abs/2111.07832

Changes from original:
* Per epoch weight decay scheduling (instead of per step)
"""

# Reference license: Apache-2.0

import argparse
import json
import logging
import math
import sys
import time
from collections.abc import Callable
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.amp
import torchinfo
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from torchvision.datasets.folder import pil_loader  # Slower but Handles external dataset quirks better
from torchvision.transforms import v2
from tqdm import tqdm

import birder
from birder.common import cli
from birder.common import fs_ops
from birder.common import training_cli
from birder.common import training_utils
from birder.common.lib import format_duration
from birder.common.lib import get_mim_network_name
from birder.common.lib import get_network_name
from birder.common.masking import BlockMasking
from birder.conf import settings
from birder.data.dataloader.webdataset import make_wds_loader
from birder.data.datasets.directory import make_image_dataset
from birder.data.datasets.directory import tv_loader
from birder.data.datasets.fake import FakeDataWithPaths
from birder.data.datasets.webdataset import make_wds_dataset
from birder.data.datasets.webdataset import prepare_wds_args
from birder.data.datasets.webdataset import wds_args_from_info
from birder.data.transforms.classification import RGBType
from birder.data.transforms.classification import get_rgb_stats
from birder.model_registry import Task
from birder.model_registry import registry
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import get_signature
from birder.net.ssl.base import get_ssl_signature
from birder.net.ssl.ibot import iBOT
from birder.net.ssl.ibot import iBOTLoss

logger = logging.getLogger(__name__)


class TrainTransform:
    def __init__(
        self,
        global_transform: Callable[..., torch.Tensor],
        crop_size: tuple[int, int],
        rgv_values: RGBType,
        local_crops_number: int,
        mask_generator: Callable[[int], torch.Tensor],
    ) -> None:
        self.global_transform = global_transform
        self.local_crops_number = local_crops_number

        # Local small crops
        mean = rgv_values["mean"]
        std = rgv_values["std"]
        self.local_transform = v2.Compose(
            [
                v2.PILToTensor(),
                v2.RandomResizedCrop(crop_size, scale=(0.1, 0.5), interpolation=v2.InterpolationMode.BICUBIC),
                v2.RandomHorizontalFlip(p=0.5),
                v2.RandomApply([v2.ColorJitter(brightness=0.25, contrast=0.15, hue=0.04)], p=0.8),
                v2.RandomApply([v2.GaussianBlur(kernel_size=(3, 3), sigma=(0.5, 1.2))], p=0.5),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=mean, std=std),
            ]
        )
        self.mask_generator = mask_generator

    def __call__(self, image: Any) -> tuple[list[torch.Tensor], torch.Tensor]:
        crops = []
        crops.append(self.global_transform(image))
        crops.append(self.global_transform(image))
        for _ in range(self.local_crops_number):
            crops.append(self.local_transform(image))

        masks = self.mask_generator(2)

        return (crops, masks)


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def train(args: argparse.Namespace) -> None:
    #
    # Initialize
    #
    device, device_id, disable_tqdm = training_utils.init_training(args, logger)

    if args.size is None:
        args.size = registry.get_default_size(args.network)

    logger.info(f"Using size={args.size}")

    batch_size: int = args.batch_size
    grad_accum_steps: int = args.grad_accum_steps
    logger.debug(f"Effective batch size = {batch_size * grad_accum_steps * args.world_size}")

    begin_epoch = 1
    epochs = args.epochs + 1
    if args.stop_epoch is None:
        args.stop_epoch = epochs
    else:
        args.stop_epoch += 1

    #
    # Initialize network
    #
    model_dtype: torch.dtype = getattr(torch, args.model_dtype)
    sample_shape = (batch_size, args.channels, *args.size)  # B, C, H, W
    backbone_name = get_network_name(args.network, tag="ibot")
    if args.tag is not None:
        backbone_name = f"{backbone_name}-{args.tag}"

    network_name = get_mim_network_name("ibot", encoder=args.network, tag=args.tag)

    student_backbone = registry.net_factory(args.network, 0, sample_shape[1], config=args.model_config, size=args.size)
    if args.model_config is not None:
        teacher_model_config = args.model_config.copy()
        teacher_model_config.update({"drop_path_rate": 0.0})
    else:
        teacher_model_config = {"drop_path_rate": 0.0}

    teacher_backbone = registry.net_factory(
        args.network, 0, sample_shape[1], config=teacher_model_config, size=args.size
    )
    student_backbone.set_dynamic_size()
    student = iBOT(
        student_backbone,
        config={
            "out_dim": args.out_dim,
            "norm_last_layer": args.norm_last_layer,
            "num_layers": 3,
            "hidden_dim": 2048,
            "bottleneck_dim": 256,
            "patch_out_dim": args.patch_out_dim,
            "shared_head": args.shared_head,
        },
    )
    teacher = iBOT(
        teacher_backbone,
        config={
            "out_dim": args.out_dim,
            "norm_last_layer": True,
            "num_layers": 3,
            "hidden_dim": 2048,
            "bottleneck_dim": 256,
            "patch_out_dim": args.patch_out_dim,
            "shared_head": args.shared_head,
        },
    )
    teacher.load_state_dict(student.state_dict())
    teacher.eval()

    ibot_loss = iBOTLoss(
        args.out_dim,
        args.patch_out_dim,
        num_global_crops=2,
        num_local_crops=args.local_crops_number,
        warmup_teacher_temp=args.warmup_teacher_temp,
        teacher_temp=args.teacher_temp,
        warmup_teacher_temp2=args.warmup_teacher_temp,
        teacher_temp2=args.teacher_temp,
        warmup_teacher_temp_epochs=args.warmup_teacher_temp_epochs,
        epochs=args.epochs,
        student_temp=0.1,
        center_momentum=0.9,
        center_momentum2=0.9,
        lambda1=args.lambda1,
        lambda2=args.lambda2,
        mim_start_epoch=args.pred_start_epoch,
    )

    net = torch.nn.ModuleDict(
        {
            "student": student,
            "teacher": teacher,
            "loss": ibot_loss,
        }
    )
    net.task = teacher.task

    if args.resume_epoch is not None:
        begin_epoch = args.resume_epoch + 1
        net, training_states = fs_ops.load_simple_checkpoint(
            device, net, network_name, epoch=args.resume_epoch, strict=not args.non_strict_weights
        )
        student = net["student"]
        teacher = net["teacher"]
        ibot_loss = net["loss"]

    else:
        training_states = fs_ops.TrainingStates.empty()

    assert isinstance(student_backbone, MaskedTokenRetentionMixin)
    assert isinstance(net, torch.nn.Module)

    net.to(device, dtype=model_dtype)
    if args.freeze_bn is True:
        student = training_utils.freeze_batchnorm2d(student)
        teacher = training_utils.freeze_batchnorm2d(teacher)
    elif args.sync_bn is True and args.distributed is True:
        student = torch.nn.SyncBatchNorm.convert_sync_batchnorm(student)
        teacher = torch.nn.SyncBatchNorm.convert_sync_batchnorm(teacher)

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    # Compile networks
    teacher_compile_flag = args.compile is True or args.compile_teacher is True
    if args.compile is True:
        student = torch.compile(student)
        teacher = torch.compile(teacher)
    elif args.compile_teacher is True:
        teacher = torch.compile(teacher)

    #
    # Data
    #
    rgb_stats = get_rgb_stats(args.rgb_mode, args.rgb_mean, args.rgb_std)
    logger.debug(f"Using RGB stats: {rgb_stats}")

    mask_size = (args.size[0] // student_backbone.max_stride, args.size[1] // student_backbone.max_stride)
    mask_generator = BlockMasking(
        mask_size,
        min_num_patches=4,
        max_num_patches=mask_size[0] * mask_size[1] // 2,
        min_aspect=0.33,
        max_aspect=3.33,
    )
    training_transform = TrainTransform(
        training_utils.get_training_transform(args),
        args.local_crop_size,
        rgb_stats,
        args.local_crops_number,
        mask_generator,
    )
    if args.use_fake_data is True:
        logger.warning("Using fake data")
        training_dataset = FakeDataWithPaths(
            10000, (args.channels, *args.size), num_classes=10, transform=training_transform
        )

    elif args.wds is True:
        wds_path: str | list[str]
        if args.wds_info is not None:
            wds_path, dataset_size = wds_args_from_info(args.wds_info, args.wds_split)
            if args.wds_size is not None:
                dataset_size = args.wds_size
        else:
            wds_path, dataset_size = prepare_wds_args(args.data_path[0], args.wds_size, device)

        training_dataset = make_wds_dataset(
            wds_path,
            dataset_size=dataset_size,
            shuffle=True,
            samples_names=True,
            transform=training_transform,
            img_loader=args.img_loader,
            cls_key=None,
            cache_dir=args.wds_cache_dir,
        )

    else:
        training_dataset = make_image_dataset(
            args.data_path,
            {},
            transforms=training_transform,
            loader=pil_loader if args.img_loader == "pil" else tv_loader,
        )

    logger.info(f"Using device {device}:{device_id}")
    logger.info(f"Training dataset has {len(training_dataset):,} samples")

    # Data loaders and samplers
    virtual_epoch_mode = args.steps_per_epoch is not None
    train_sampler, _ = training_utils.get_samplers(
        args, training_dataset, validation_dataset=None, infinite=virtual_epoch_mode
    )

    if args.wds is True:
        training_loader = make_wds_loader(
            training_dataset,
            batch_size,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            collate_fn=None,
            world_size=args.world_size,
            pin_memory=True,
            drop_last=args.drop_last,
            shuffle=args.wds_extra_shuffle,
            infinite=virtual_epoch_mode,
        )

    else:
        training_loader = DataLoader(
            training_dataset,
            batch_size=batch_size,
            sampler=train_sampler,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            pin_memory=True,
            drop_last=args.drop_last,
        )

    if virtual_epoch_mode is True:
        optimizer_steps_per_epoch = args.steps_per_epoch
        epoch_num_batches = args.steps_per_epoch * grad_accum_steps
        epoch_samples = epoch_num_batches * batch_size * args.world_size
        logger.debug(f"Virtual epoch has {epoch_samples:,} samples")
    else:
        optimizer_steps_per_epoch = math.ceil(len(training_loader) / grad_accum_steps)
        epoch_num_batches = len(training_loader)
        epoch_samples = len(training_dataset)

    last_batch_idx = epoch_num_batches - 1
    logger.debug(
        f"Epoch has {epoch_num_batches} iterations ({optimizer_steps_per_epoch} steps), "
        f"virtual mode={virtual_epoch_mode}"
    )

    #
    # Loss criteria, optimizer, learning rate scheduler and training parameter groups
    #

    # Learning rate scaling
    lr = training_utils.scale_lr(args)

    # Training parameter groups
    custom_keys_weight_decay = training_utils.get_wd_custom_keys(args)
    parameters = training_utils.optimizer_parameter_groups(
        net,
        args.wd,
        base_lr=lr,
        norm_weight_decay=args.norm_wd,
        custom_keys_weight_decay=custom_keys_weight_decay,
        custom_layer_weight_decay=args.custom_layer_wd,
        layer_decay=args.layer_decay,
        layer_decay_min_scale=args.layer_decay_min_scale,
        layer_decay_no_opt_scale=args.layer_decay_no_opt_scale,
        bias_lr=args.bias_lr,
        custom_layer_lr_scale=args.custom_layer_lr_scale,
    )

    if args.lr_scheduler_update == "epoch":
        step_update = False
        scheduler_steps_per_epoch = 1
    elif args.lr_scheduler_update == "step":
        step_update = True
        scheduler_steps_per_epoch = optimizer_steps_per_epoch
    else:
        raise ValueError("Unsupported lr_scheduler_update")

    # Optimizer and learning rate scheduler
    optimizer = training_utils.get_optimizer(parameters, lr, args)
    scheduler = training_utils.get_scheduler(optimizer, scheduler_steps_per_epoch, args)
    if args.compile_opt is True:
        optimizer.step = torch.compile(optimizer.step, fullgraph=False)

    # Teacher momentum and weight decay schedule
    momentum_schedule = training_utils.cosine_scheduler(args.momentum_teacher, 1.0, args.epochs, 0, epoch_num_batches)
    if args.wd_end is not None:
        wd_schedule = training_utils.cosine_scheduler(args.wd, args.wd_end, args.epochs, 0, 1)
    else:
        wd_schedule = None

    # Gradient scaler and AMP related tasks
    scaler, amp_dtype = training_utils.get_amp_scaler(args.amp, args.amp_dtype)

    # Load states
    if args.load_states is True:
        optimizer.load_state_dict(training_states.optimizer_state)
        scheduler.load_state_dict(training_states.scheduler_state)
        if scaler is not None:
            scaler.load_state_dict(training_states.scaler_state)

    elif args.load_scheduler is True:
        scheduler.load_state_dict(training_states.scheduler_state)
        last_lrs = scheduler.get_last_lr()
        for g, last_lr in zip(optimizer.param_groups, last_lrs):
            g["lr"] = last_lr

    last_lr = float(max(scheduler.get_last_lr()))
    if args.plot_lr is True:
        logger.info("Fast forwarding scheduler...")
        optimizer.step()
        lrs = []
        for _ in range(begin_epoch, epochs):
            for _ in range(scheduler_steps_per_epoch):
                lrs.append(float(max(scheduler.get_last_lr())))
                scheduler.step()

        plt.plot(
            np.linspace(begin_epoch, epochs, scheduler_steps_per_epoch * (epochs - begin_epoch), endpoint=False), lrs
        )
        plt.show()
        raise SystemExit(0)

    #
    # Distributed (DDP)
    #

    # There is no backpropagation through the teacher
    for p in teacher.parameters():
        p.requires_grad_(False)

    student_without_ddp = student
    if args.distributed is True:
        student = torch.nn.parallel.DistributedDataParallel(
            student, device_ids=[args.local_rank], find_unused_parameters=args.find_unused_parameters
        )
        student_without_ddp = student.module

    model_to_save = net
    if teacher_compile_flag is True and hasattr(model_to_save["teacher"], "_orig_mod") is True:
        model_to_save["teacher"] = model_to_save["teacher"]._orig_mod  # pylint: disable=protected-access
    if args.compile is True and hasattr(model_to_save["student"], "_orig_mod") is True:
        model_to_save["student"] = model_to_save["student"]._orig_mod  # pylint: disable=protected-access

    #
    # Misc
    #

    # Print network summary
    net_for_info = teacher
    if teacher_compile_flag is True and hasattr(teacher, "_orig_mod") is True:
        net_for_info = teacher._orig_mod  # pylint: disable=protected-access

    if args.no_summary is False:
        summary = torchinfo.summary(
            net_for_info,
            device=device,
            input_data={"x": torch.rand(sample_shape), "masks": mask_generator(batch_size)},
            dtypes=[model_dtype],
            col_names=["input_size", "output_size", "kernel_size", "num_params"],
            depth=4,
            verbose=0,
        )
        if training_utils.is_global_primary(args) is True:
            # Write to stderr, same as all the logs
            print(summary, file=sys.stderr)

    # Training logs
    training_log_path = training_utils.training_log_path(network_name, device, args.experiment)
    logger.info(f"Logging training run at {training_log_path}")
    summary_writer = SummaryWriter(training_log_path)

    signature = get_ssl_signature(input_shape=sample_shape)
    backbone_signature = get_signature(input_shape=sample_shape, num_outputs=0)
    file_handler: logging.Handler = logging.NullHandler()
    if training_utils.is_local_primary(args) is True:
        summary_writer.flush()
        fs_ops.write_config(network_name, net_for_info, signature=signature, rgb_stats=rgb_stats)
        file_handler = training_utils.setup_file_logging(training_log_path.joinpath("training.log"))
        with open(training_log_path.joinpath("training_args.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {
                    "birder_version": birder.__version__,
                    "pytorch_version": torch.__version__,
                    "cmdline": " ".join(sys.argv),
                    **vars(args),
                },
                handle,
                indent=2,
            )

        with open(training_log_path.joinpath("training_data.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {"training_samples": len(training_dataset)},
                handle,
                indent=2,
            )

    #
    # Training loop
    #
    if virtual_epoch_mode is True:
        train_iter = iter(training_loader)

    running_loss = training_utils.SmoothedValue()
    train_proto_agreement = training_utils.SmoothedValue()

    logger.info(f"Starting training with learning rate of {last_lr}")
    for epoch in range(begin_epoch, args.stop_epoch):
        tic = time.time()
        net.train()

        # Clear metrics
        running_loss.clear()
        train_proto_agreement.clear()

        if args.distributed is True or virtual_epoch_mode is True:
            train_sampler.set_epoch(epoch)

        if wd_schedule is not None:
            wd = wd_schedule[epoch - 1]
            for param_group in optimizer.param_groups:
                if param_group["weight_decay"] > 0:
                    param_group["weight_decay"] = wd

            logger.info(f"Updated wd to: {wd}")

        progress = tqdm(
            desc=f"Epoch {epoch}/{epochs-1}",
            total=epoch_samples,
            leave=False,
            disable=disable_tqdm,
            unit="samples",
            initial=0,
        )

        # Zero the parameter gradients
        optimizer.zero_grad()

        epoch_start = time.time()
        start_time = epoch_start
        last_idx = -1
        batch_iter: Iterator[tuple[int, Any]]
        if virtual_epoch_mode is True:
            batch_iter = ((i, next(train_iter)) for i in range(epoch_num_batches))
        else:
            batch_iter = enumerate(training_loader)

        for i, (_, (images, masks), _) in batch_iter:
            global_iter = ((epoch - 1) * epoch_num_batches) + i
            images = [img.to(device, dtype=model_dtype, non_blocking=True) for img in images]
            if args.pred_start_epoch >= epoch:
                masks = None
            else:
                masks = masks.to(device, dtype=model_dtype, non_blocking=True)

            optimizer_update = (i == last_batch_idx) or ((i + 1) % grad_accum_steps == 0)

            # Forward, backward and optimize
            with torch.amp.autocast("cuda", enabled=args.amp, dtype=amp_dtype):
                # Global views
                with torch.no_grad():
                    teacher_embedding, teacher_features = teacher(torch.concat(images[:2], dim=0), None)

                student_embedding, student_features = student(torch.concat(images[:2], dim=0), masks)

                # Local views
                student_local_embedding, _ = student(torch.concat(images[2:], dim=0), None, return_keys="embedding")

                loss = ibot_loss(
                    student_embedding,
                    student_features,
                    teacher_embedding,
                    teacher_features,
                    student_local_embedding,
                    masks.reshape(batch_size * 2, -1),
                    epoch - 1,
                )["all"]

            if scaler is not None:
                scaler.scale(loss).backward()
                if args.freeze_last_layer_epochs >= epoch:
                    student_without_ddp.head.cancel_last_layer_gradients()

                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        scaler.unscale_(optimizer)
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    scaler.step(optimizer)
                    scaler.update()
                    optimizer.zero_grad()
                    if step_update is True:
                        scheduler.step()

            else:
                loss.backward()
                if args.freeze_last_layer_epochs >= epoch:
                    student_without_ddp.head.cancel_last_layer_gradients()

                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    optimizer.step()
                    optimizer.zero_grad()
                    if step_update is True:
                        scheduler.step()

            if optimizer_update is True:
                # EMA update for the teacher
                with torch.no_grad():
                    m = momentum_schedule[global_iter]
                    torch._foreach_lerp_(  # pylint: disable=protected-access
                        list(teacher.parameters()), list(student.parameters()), weight=1 - m
                    )

            # Statistics
            running_loss.update(loss.detach())

            probs_teacher = teacher_embedding.chunk(2)  # Per global crop
            probs_student = student_embedding.chunk(2)
            pred_teacher = probs_teacher[0].argmax(dim=1)
            pred_student = probs_student[1].argmax(dim=1)
            train_proto_agreement.update(training_utils.accuracy(pred_teacher, pred_student))

            # Write statistics
            if (i % args.log_interval == 0 and i > 0) or i == last_batch_idx:
                time_now = time.time()
                time_cost = time_now - start_time
                iters_processed_in_interval = i - last_idx
                rate = iters_processed_in_interval * (batch_size * args.world_size) / time_cost

                avg_time_per_iter = time_cost / iters_processed_in_interval
                remaining_iters_in_epoch = last_batch_idx - i
                estimated_time_to_finish_epoch = remaining_iters_in_epoch * avg_time_per_iter

                start_time = time_now
                last_idx = i
                cur_lr = float(max(scheduler.get_last_lr()))

                running_loss.synchronize_between_processes(device)
                train_proto_agreement.synchronize_between_processes(device)
                with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
                    log.info(
                        f"[Trn] Epoch {epoch}/{epochs-1}, iter {i+1}/{last_batch_idx+1}  "
                        f"Loss: {running_loss.avg:.4f}  "
                        f"Elapsed: {format_duration(time_now-epoch_start)}  "
                        f"ETA: {format_duration(estimated_time_to_finish_epoch)}  "
                        f"T: {time_cost:.1f}s  "
                        f"R: {rate:.1f} samples/s  "
                        f"LR: {cur_lr:.4e}"
                    )

                if training_utils.is_local_primary(args) is True:
                    summary_writer.add_scalars(
                        "loss",
                        {"training": running_loss.avg},
                        ((epoch - 1) * epoch_samples) + ((i + 1) * batch_size * args.world_size),
                    )
                    summary_writer.add_scalars(
                        "performance",
                        {"prototype_agreement": train_proto_agreement.avg},
                        ((epoch - 1) * epoch_samples) + ((i + 1) * batch_size * args.world_size),
                    )

            # Update progress bar
            progress.update(n=batch_size * args.world_size)

        progress.close()

        # Epoch training metrics
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} training_loss: {running_loss.global_avg:.4f}")
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} prototype_agreement: {train_proto_agreement.global_avg:.4f}")

        # Learning rate scheduler update
        if step_update is False:
            scheduler.step()
        if last_lr != float(max(scheduler.get_last_lr())):
            last_lr = float(max(scheduler.get_last_lr()))
            logger.info(f"Updated learning rate to: {last_lr}")

        if training_utils.is_local_primary(args) is True:
            # Checkpoint model
            if epoch % args.save_frequency == 0:
                fs_ops.checkpoint_model(
                    network_name,
                    epoch,
                    model_to_save,
                    signature,
                    {},
                    rgb_stats,
                    optimizer,
                    scheduler,
                    scaler,
                    None,
                )
                fs_ops.checkpoint_model(
                    backbone_name,
                    epoch,
                    model_to_save["teacher"].backbone,
                    backbone_signature,
                    {},
                    rgb_stats,
                    optimizer=None,
                    scheduler=None,
                    scaler=None,
                    model_base=None,
                )
                if args.keep_last is not None:
                    fs_ops.clean_checkpoints(network_name, args.keep_last)
                    fs_ops.clean_checkpoints(backbone_name, args.keep_last)

        # Epoch timing
        toc = time.time()
        logger.info(f"Total time: {format_duration(toc - tic)}")
        logger.info("---")

    summary_writer.close()

    # Checkpoint model
    if training_utils.is_local_primary(args) is True:
        fs_ops.checkpoint_model(
            network_name,
            epoch,
            model_to_save,
            signature,
            {},
            rgb_stats,
            optimizer,
            scheduler,
            scaler,
            None,
        )
        fs_ops.checkpoint_model(
            backbone_name,
            epoch,
            model_to_save["teacher"].backbone,
            backbone_signature,
            {},
            rgb_stats,
            optimizer=None,
            scheduler=None,
            scaler=None,
            model_base=None,
        )

    training_utils.shutdown_distributed_mode(args)


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Pre-train model",
        epilog=(
            "Usage examples\n"
            "==============\n"
            "torchrun --nproc_per_node=2 -m birder.scripts.train_ibot \\\n"
            "    --network vit_b16 \\\n"
            "    --shared-head \\\n"
            "    --norm-last-layer \\\n"
            "    --local-crops-number 10 \\\n"
            "    --teacher-temp 0.07 \\\n"
            "    --warmup-teacher-temp-epochs 50 \\\n"
            "    --opt adamw \\\n"
            "    --lr 0.00075 \\\n"
            "    --lr-scheduler cosine \\\n"
            "    --lr-cosine-min 1e-6 \\\n"
            "    --freeze-last-layer-epochs 3 \\\n"
            "    --epochs 400 \\\n"
            "    --warmup-epochs 10 \\\n"
            "    --batch-size 32 \\\n"
            "    --wd 0.04 \\\n"
            "    --wd-end 0.4 \\\n"
            "    --norm-wd 0 \\\n"
            "    --bias-weight-decay 0 \\\n"
            "    --clip-grad-norm 0.3 \\\n"
            "    --amp \\\n"
            "    --compile \\\n"
            "    --data-path data/training\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("-n", "--network", type=str, help="the neural network to use")
    parser.add_argument("-t", "--tag", type=str, help="add model tag")
    parser.add_argument(
        "--model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the model default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("--out-dim", type=int, default=8192, help="dimensionality of embedding output")
    parser.add_argument("--patch-out-dim", type=int, default=8192, help="dimensionality of features output")
    parser.add_argument(
        "--shared-head",
        default=False,
        action="store_true",
        help="wether to share the same head for embedding and features outputs (if false patch_out_dim is ignored)",
    )
    parser.add_argument(
        "--norm-last-layer",
        default=False,
        action="store_true",
        help="whether or not to weight normalize the last layer of the head, set this flag with large models (vit_b)",
    )
    parser.add_argument(
        "--momentum-teacher",
        type=float,
        default=0.996,
        help="base EMA parameter for teacher update, set a higher value with small batches",
    )
    parser.add_argument("--local-crops-number", type=int, default=8, help="number of small local views to generate")
    parser.add_argument(
        "--local-crop-size", type=int, nargs="+", default=[96, 96], metavar=("H", "W"), help="local view size"
    )
    parser.add_argument(
        "--pred-start-epoch", type=int, default=0, help="start epoch to perform masked image prediction"
    )
    parser.add_argument("--lambda1", default=1.0, type=float, help="loss weight for dino loss over embedding")
    parser.add_argument("--lambda2", default=1.0, type=float, help="loss weight for beit loss over features")
    parser.add_argument(
        "--warmup-teacher-temp",
        type=float,
        default=0.04,
        help="initial value for the teacher temperature, try decreasing it if the training loss does not decrease",
    )
    parser.add_argument(
        "--teacher-temp", type=float, default=0.04, help="final value (after linear warmup) of the teacher temperature"
    )
    parser.add_argument(
        "--warmup-teacher-temp-epochs", type=int, default=40, help="number of warmup epochs for the teacher temperature"
    )
    parser.add_argument(
        "--freeze-last-layer-epochs",
        default=1,
        type=int,
        help=(
            "number of epochs during which the output layer is frozen, "
            "try increasing this value if the loss does not decrease"
        ),
    )
    training_cli.add_optimization_args(parser)
    training_cli.add_lr_wd_args(parser, wd_end=True)
    training_cli.add_lr_scheduler_args(parser)
    training_cli.add_training_schedule_args(parser, default_epochs=800)
    training_cli.add_batch_norm_args(parser)
    training_cli.add_input_args(parser)
    training_cli.add_data_aug_args(parser, default_level=1, default_min_scale=0.3, default_re_prob=0.0)
    training_cli.add_dataloader_args(parser, default_drop_last=True)
    training_cli.add_precision_args(parser)
    training_cli.add_compile_args(parser, teacher=True)
    training_cli.add_checkpoint_args(parser)
    training_cli.add_distributed_args(parser)
    training_cli.add_logging_and_debug_args(parser, default_log_interval=100)
    training_cli.add_training_data_args(parser, unsupervised=True)

    return parser


def validate_args(args: argparse.Namespace) -> None:
    args.data_path = [str(p) for p in args.data_path]
    args.size = cli.parse_size(args.size)
    args.local_crop_size = cli.parse_size(args.local_crop_size)

    # This will capture the common argument mistakes
    training_cli.common_args_validation(args)

    # Script specific checks
    if registry.exists(args.network, task=Task.IMAGE_CLASSIFICATION, net_type=MaskedTokenRetentionMixin) is False:
        raise cli.ValidationError(f"--network {args.network} not supported, see list-models tool for available options")


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.MODELS_DIR.exists() is False:
        logger.info(f"Creating {settings.MODELS_DIR} directory...")
        settings.MODELS_DIR.mkdir(parents=True, exist_ok=True)

    if args.wds_cache_dir is not None and Path(args.wds_cache_dir).exists() is False:
        logger.info(f"Creating {args.wds_cache_dir} directory...")
        Path(args.wds_cache_dir).mkdir(parents=True, exist_ok=True)

    train(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()



"""
Guided Backpropagation, adapted from
https://github.com/jacobgil/pytorch-grad-cam

Paper "Striving for Simplicity: The All Convolutional Net", https://arxiv.org/abs/1412.6806
"""

# Reference license: MIT

import math
from collections.abc import Callable
from pathlib import Path
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from PIL import Image
from torch import nn
from torch.autograd import Function

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import deprocess_image
from birder.introspection.base import predict_class
from birder.introspection.base import preprocess_image
from birder.introspection.base import validate_target_class


# pylint: disable=abstract-method,arguments-differ
class GuidedBackpropReLU(Function):
    @staticmethod
    def forward(ctx: Any, input_img: torch.Tensor) -> torch.Tensor:
        positive_mask = (input_img > 0).type_as(input_img)
        output = torch.addcmul(torch.zeros(input_img.size()).type_as(input_img), input_img, positive_mask)
        ctx.save_for_backward(input_img, output)

        return output

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> torch.Tensor:
        input_img, _output = ctx.saved_tensors

        positive_mask_1 = (input_img > 0).type_as(grad_output)
        positive_mask_2 = (grad_output > 0).type_as(grad_output)
        grad_input = torch.addcmul(
            torch.zeros(input_img.size()).type_as(input_img),
            torch.addcmul(torch.zeros(input_img.size()).type_as(input_img), grad_output, positive_mask_1),
            positive_mask_2,
        )

        return grad_input


# pylint: disable=abstract-method,arguments-differ
class GuidedBackpropSiLU(Function):
    @staticmethod
    def forward(ctx: Any, input_img: torch.Tensor) -> torch.Tensor:
        result = input_img * torch.sigmoid(input_img)
        ctx.save_for_backward(input_img)

        return result

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> torch.Tensor:
        i = ctx.saved_tensors[0]
        sigmoid_i = torch.sigmoid(i)
        positive_mask_1 = (i > 0).type_as(grad_output)
        positive_mask_2 = (grad_output > 0).type_as(grad_output)
        grad_input = grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i))) * positive_mask_1 * positive_mask_2

        return grad_input


# pylint: disable=abstract-method,arguments-differ
class GuidedBackpropGELU(Function):
    @staticmethod
    def forward(ctx: Any, input_img: torch.Tensor) -> torch.Tensor:
        result = F.gelu(input_img, approximate="none")  # pylint:disable=not-callable
        ctx.save_for_backward(input_img)
        return result

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> torch.Tensor:
        x = ctx.saved_tensors[0]

        sqrt_2 = math.sqrt(2.0)
        sqrt_2pi = math.sqrt(2.0 * math.pi)

        cdf = 0.5 * (1.0 + torch.erf(x / sqrt_2))
        pdf = torch.exp(-0.5 * x * x) / sqrt_2pi

        d_gelu = cdf + x * pdf

        positive_mask_1 = (x > 0).type_as(grad_output)
        positive_mask_2 = (grad_output > 0).type_as(grad_output)

        grad_input = grad_output * d_gelu * positive_mask_1 * positive_mask_2

        return grad_input


# pylint: disable=abstract-method,arguments-differ
class GuidedBackpropHardswish(Function):
    @staticmethod
    def forward(ctx: Any, input_img: torch.Tensor) -> torch.Tensor:
        result = F.hardswish(input_img)
        ctx.save_for_backward(input_img)
        return result

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> torch.Tensor:
        x = ctx.saved_tensors[0]

        grad = torch.zeros_like(x)

        mask_mid = (x > -3) & (x < 3)
        grad[mask_mid] = (2.0 * x[mask_mid] + 3.0) / 6.0

        mask_high = x >= 3
        grad[mask_high] = 1.0

        positive_mask_1 = (x > 0).type_as(grad_output)
        positive_mask_2 = (grad_output > 0).type_as(grad_output)

        grad_input = grad_output * grad * positive_mask_1 * positive_mask_2

        return grad_input


class GuidedReLU(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return GuidedBackpropReLU.apply(x)


class GuidedSiLU(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return GuidedBackpropSiLU.apply(x)


class GuidedGELU(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return GuidedBackpropGELU.apply(x)


class GuidedHardswish(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return GuidedBackpropHardswish.apply(x)


# Activation replacement mapping
ACTIVATION_REPLACEMENTS: dict[type, type] = {
    nn.ReLU: GuidedReLU,
    nn.SiLU: GuidedSiLU,
    nn.GELU: GuidedGELU,
    nn.Hardswish: GuidedHardswish,
}


def replace_activations_recursive(model: nn.Module, replacements: dict[type, type]) -> None:
    """
    NOTE: This ONLY works for activations defined as nn.Module objects (e.g., self.act = nn.ReLU()).
    It will NOT affect functional calls inside forward methods, such as F.relu(x) or F.gelu(x).
    """

    for name, module in list(model._modules.items()):  # pylint: disable=protected-access
        for old_type, new_type in replacements.items():
            if isinstance(module, old_type):
                model._modules[name] = new_type()  # pylint: disable=protected-access
                break
        else:
            # Recurse into submodules
            replace_activations_recursive(module, replacements)


def restore_activations_recursive(model: nn.Module, guided_types: dict[type, type]) -> None:
    reverse_mapping = {v: k for k, v in guided_types.items()}
    for name, module in list(model._modules.items()):  # pylint: disable=protected-access
        for guided_type, original_type in reverse_mapping.items():
            if isinstance(module, guided_type):
                model._modules[name] = original_type()  # pylint: disable=protected-access
                break
        else:
            restore_activations_recursive(module, guided_types)


class GuidedBackprop:
    def __init__(self, net: nn.Module, device: torch.device, transform: Callable[..., torch.Tensor]) -> None:
        self.net = net.eval()
        self.device = device
        self.transform = transform

    def __call__(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        input_tensor, rgb_img = preprocess_image(image, self.transform, self.device)

        # Get prediction
        with torch.inference_mode():
            logits = self.net(input_tensor)

        if target_class is None:
            target_class = predict_class(logits)
        else:
            validate_target_class(target_class, logits.shape[-1])

        # Replace activations with guided versions
        replace_activations_recursive(self.net, ACTIVATION_REPLACEMENTS)

        try:
            input_tensor = input_tensor.detach().requires_grad_(True)
            output = self.net(input_tensor)

            loss = output[0, target_class]
            loss.backward(retain_graph=False)

            gradients = input_tensor.grad.cpu().numpy()
            gradients = gradients[0, :, :, :]  # Remove batch dim
            gradients = gradients.transpose((1, 2, 0))  # CHW -> HWC

        finally:
            restore_activations_recursive(self.net, ACTIVATION_REPLACEMENTS)

        visualization = deprocess_image(gradients * rgb_img)

        return InterpretabilityResult(
            original_image=rgb_img,
            visualization=visualization,
            raw_output=gradients,
            logits=logits.detach(),
            predicted_class=target_class,
        )



"""
Plankton benchmark using linear probing for phytoplankton classification

Dataset "SYKE-plankton_IFCB_2022", https://b2share.eudat.eu/records/xvnrp-7ga56
"""

import argparse
import logging
import os
import time
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
import polars as pl
import torch
from rich.console import Console
from rich.table import Table

from birder.common import cli
from birder.common import lib
from birder.conf import settings
from birder.data.datasets.directory import make_image_dataset
from birder.datahub.evaluation import Plankton
from birder.eval._embeddings import load_embeddings
from birder.eval.methods.linear import evaluate_linear_probe
from birder.eval.methods.linear import train_linear_probe

logger = logging.getLogger(__name__)


def _print_summary_table(results: list[dict[str, Any]]) -> None:
    console = Console()

    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("Plankton (Linear)", style="dim")
    table.add_column("Accuracy", justify="right")
    table.add_column("Std", justify="right")
    table.add_column("Runs", justify="right")

    for result in results:
        table.add_row(
            Path(result["embeddings_file"]).name,
            f"{result['accuracy']:.4f}",
            f"{result['accuracy_std']:.4f}",
            f"{result['num_runs']}",
        )

    console.print(table)


def _write_results_csv(results: list[dict[str, Any]], output_path: Path) -> None:
    rows: list[dict[str, Any]] = []
    for result in results:
        rows.append(
            {
                "embeddings_file": result["embeddings_file"],
                "method": result["method"],
                "accuracy": result["accuracy"],
                "accuracy_std": result["accuracy_std"],
                "num_runs": result["num_runs"],
            }
        )

    pl.DataFrame(rows).write_csv(output_path)
    logger.info(f"Results saved to {output_path}")


def _load_plankton_metadata(dataset: Plankton) -> pl.DataFrame:
    """
    Load metadata using make_image_dataset with a fixed class_to_idx

    Returns DataFrame with columns: id (filename stem), label, split
    """

    # Build class_to_idx from train classes only
    class_to_idx = {
        entry.name: idx
        for idx, entry in enumerate(sorted(os.scandir(str(dataset.train_dir)), key=lambda e: e.name))
        if entry.is_dir()
    }

    rows: list[dict[str, Any]] = []
    skipped_no_label = 0
    for split, split_dir in [("train", dataset.train_dir), ("val", dataset.val_dir)]:
        image_dataset = make_image_dataset([str(split_dir)], class_to_idx)
        for i in range(len(image_dataset)):
            path = image_dataset.paths[i].decode("utf-8")
            label = image_dataset.labels[i].item()
            if label == settings.NO_LABEL:
                skipped_no_label += 1
                continue

            rows.append({"id": Path(path).stem, "label": label, "split": split})

    if skipped_no_label > 0:
        logger.info(f"Skipped {skipped_no_label} samples with unknown labels (NO_LABEL)")

    return pl.DataFrame(rows)


def _load_embeddings_with_split(
    embeddings_path: str, metadata_df: pl.DataFrame
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.int_], npt.NDArray[np.float32], npt.NDArray[np.int_]]:
    logger.info(f"Loading embeddings from {embeddings_path}")
    emb_df = load_embeddings(embeddings_path)

    train_meta = metadata_df.filter(pl.col("split") == "train").select(["id", "label"])
    val_meta = metadata_df.filter(pl.col("split") == "val").select(["id", "label"])

    train_join = train_meta.join(emb_df, on="id", how="inner")
    val_join = val_meta.join(emb_df, on="id", how="inner")

    dropped_train = train_meta.height - train_join.height
    dropped_val = val_meta.height - val_join.height
    dropped_total = dropped_train + dropped_val
    if dropped_total > 0:
        logger.warning(
            f"Join dropped {dropped_total} samples (missing embeddings): train={dropped_train}, val={dropped_val}"
        )

    x_train = train_join.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    y_train = train_join.get_column("label").to_numpy().astype(np.int_)
    x_val = val_join.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    y_val = val_join.get_column("label").to_numpy().astype(np.int_)

    num_classes = y_train.max() + 1
    total_samples = x_train.shape[0] + x_val.shape[0]
    embedding_dim = x_train.shape[1]
    logger.info(f"Loaded {total_samples} samples with {embedding_dim} dimensions, {num_classes} classes")

    logger.info(f"Train: {len(y_train)} samples, Val: {len(y_val)} samples")

    return (x_train, y_train, x_val, y_val)


def evaluate_plankton_single(
    x_train: npt.NDArray[np.float32],
    y_train: npt.NDArray[np.int_],
    x_val: npt.NDArray[np.float32],
    y_val: npt.NDArray[np.int_],
    args: argparse.Namespace,
    embeddings_path: str,
    device: torch.device,
) -> dict[str, Any]:
    num_classes = int(y_train.max() + 1)

    scores: list[float] = []
    for run in range(args.runs):
        run_seed = args.seed + run
        logger.info(f"Run {run + 1}/{args.runs} (seed={run_seed})")

        model = train_linear_probe(
            x_train,
            y_train,
            num_classes,
            device=device,
            epochs=args.epochs,
            batch_size=args.batch_size,
            lr=args.lr,
            seed=run_seed,
        )

        y_pred, y_true = evaluate_linear_probe(model, x_val, y_val, batch_size=args.batch_size, device=device)
        acc = float(np.mean(y_pred == y_true))
        scores.append(acc)
        logger.info(f"Run {run + 1}/{args.runs} - Accuracy: {acc:.4f}")

    scores_arr = np.array(scores)
    mean_acc = float(scores_arr.mean())
    std_acc = float(scores_arr.std(ddof=1)) if len(scores) > 1 else 0.0

    logger.info(f"Mean accuracy over {args.runs} runs: {mean_acc:.4f} +/- {std_acc:.4f} (std)")

    return {
        "method": "linear",
        "accuracy": mean_acc,
        "accuracy_std": std_acc,
        "num_runs": args.runs,
        "embeddings_file": str(embeddings_path),
    }


def evaluate_plankton(args: argparse.Namespace) -> None:
    tic = time.time()

    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")
    logger.info(f"Loading Plankton dataset from {args.dataset_path}")
    dataset = Plankton(args.dataset_path)
    metadata_df = _load_plankton_metadata(dataset)
    logger.info(f"Loaded metadata for {metadata_df.height} images")

    results: list[dict[str, Any]] = []
    total = len(args.embeddings)
    for idx, embeddings_path in enumerate(args.embeddings, start=1):
        logger.info(f"Processing embeddings {idx}/{total}: {embeddings_path}")
        x_train, y_train, x_val, y_val = _load_embeddings_with_split(embeddings_path, metadata_df)

        result = evaluate_plankton_single(x_train, y_train, x_val, y_val, args, embeddings_path, device)
        results.append(result)

    _print_summary_table(results)

    if args.dry_run is False:
        output_dir = settings.RESULTS_DIR.joinpath(args.dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        output_path = output_dir.joinpath("plankton.csv")
        _write_results_csv(results, output_path)

    toc = time.time()
    logger.info(f"Plankton benchmark completed in {lib.format_duration(toc - tic)}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "plankton",
        allow_abbrev=False,
        help="run Plankton benchmark - 50 class phytoplankton classification using linear probing",
        description="run Plankton benchmark - 50 class phytoplankton classification using linear probing",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval plankton --embeddings "
            "results/plankton_embeddings.parquet --dataset-path ~/Datasets/plankton --dry-run\n"
            "python -m birder.eval plankton --embeddings results/plankton/*.parquet "
            "--dataset-path ~/Datasets/plankton --gpu\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--embeddings", type=str, nargs="+", metavar="FILE", help="paths to embeddings parquet files"
    )
    subparser.add_argument("--dataset-path", type=str, metavar="PATH", help="path to Plankton dataset root")
    subparser.add_argument("--runs", type=int, default=3, help="number of evaluation runs")
    subparser.add_argument("--epochs", type=int, default=50, help="training epochs per run")
    subparser.add_argument("--batch-size", type=int, default=64, help="batch size for training and inference")
    subparser.add_argument("--lr", type=float, default=2e-3, help="learning rate")
    subparser.add_argument("--seed", type=int, default=0, help="base random seed")
    subparser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    subparser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    subparser.add_argument(
        "--dir", type=str, default="plankton", help="place all outputs in a sub-directory (relative to results)"
    )
    subparser.add_argument("--dry-run", default=False, action="store_true", help="skip saving results to file")
    subparser.set_defaults(func=main)


def validate_args(args: argparse.Namespace) -> None:
    if args.embeddings is None:
        raise cli.ValidationError("--embeddings is required")
    if args.dataset_path is None:
        raise cli.ValidationError("--dataset-path is required")


def main(args: argparse.Namespace) -> None:
    validate_args(args)
    evaluate_plankton(args)



"""
YOLO v4 Tiny, adapted from
https://github.com/Tianxiaomo/pytorch-YOLOv4

Paper "Scaled-YOLOv4: Scaling Cross Stage Partial Network", https://arxiv.org/abs/2011.08036
"""

# Reference license: Apache-2.0

from functools import partial
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection._yolo_anchors import resolve_anchor_groups
from birder.net.detection.yolo_v3 import YOLOAnchorGenerator
from birder.net.detection.yolo_v3 import YOLOHead
from birder.net.detection.yolo_v4 import YOLO_v4

# Scale factors per detection scale to eliminate grid sensitivity
DEFAULT_SCALE_XY = [1.05, 1.05]  # [medium, large]


class YOLOTinyNeck(nn.Module):
    def __init__(self, in_channels: list[int]) -> None:
        super().__init__()
        c4, c5 = in_channels

        self.conv_c5 = Conv2dNormActivation(
            c5,
            c5 // 2,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=partial(nn.LeakyReLU, negative_slope=0.1),
        )
        self.conv_c5_out = Conv2dNormActivation(
            c5 // 2,
            c5,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            activation_layer=partial(nn.LeakyReLU, negative_slope=0.1),
        )
        self.conv_upsample = Conv2dNormActivation(
            c5 // 2,
            c5 // 4,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=partial(nn.LeakyReLU, negative_slope=0.1),
        )
        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")

        concat_channels = c4 + c5 // 4
        self.conv_c4_out = Conv2dNormActivation(
            concat_channels,
            c4,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            activation_layer=partial(nn.LeakyReLU, negative_slope=0.1),
        )

        self.out_channels = [c4, c5]

    def forward(self, feature_list: list[torch.Tensor]) -> list[torch.Tensor]:
        c4, c5 = feature_list[-2:]

        p5 = self.conv_c5(c5)
        out_large = self.conv_c5_out(p5)

        p5_up = self.upsample(self.conv_upsample(p5))
        p4_cat = torch.concat([c4, p5_up], dim=1)

        out_medium = self.conv_c4_out(p4_cat)

        return [out_medium, out_large]


# pylint: disable=invalid-name
class YOLO_v4_Tiny(YOLO_v4):
    default_size = (416, 416)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        # self.num_classes = self.num_classes - 1 (Subtracted at parent)

        score_thresh = 0.05
        nms_thresh = 0.45
        detections_per_img = 300
        ignore_thresh = 0.7
        noobj_coeff = 0.25
        coord_coeff = 3.0
        obj_coeff = 1.0
        cls_coeff = 1.0
        label_smoothing = 0.1
        anchor_spec = self.config["anchors"]

        self.ignore_thresh = ignore_thresh
        self.noobj_coeff = noobj_coeff
        self.coord_coeff = coord_coeff
        self.obj_coeff = obj_coeff
        self.cls_coeff = cls_coeff
        self.scale_xy = DEFAULT_SCALE_XY
        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.detections_per_img = detections_per_img

        self.backbone.return_channels = self.backbone.return_channels[-2:]
        self.backbone.return_stages = self.backbone.return_stages[-2:]

        self.label_smoothing = label_smoothing
        self.smooth_positive = 1.0 - self.label_smoothing
        self.smooth_negative = self.label_smoothing / self.num_classes

        self.neck = YOLOTinyNeck(self.backbone.return_channels)

        anchors = resolve_anchor_groups(
            anchor_spec, anchor_format="pixels", model_size=self.size, model_strides=(16, 32)
        )
        self.anchor_generator = YOLOAnchorGenerator(anchors)
        num_anchors = self.anchor_generator.num_anchors_per_location()
        self.head = YOLOHead(self.neck.out_channels, num_anchors, self.num_classes)


registry.register_model_config("yolo_v4_tiny", YOLO_v4_Tiny, config={"anchors": "yolo_v4_tiny"})



import logging
import unittest

import torch
import torch.nn.functional as F

from birder.common import masking
from birder.conf.settings import DEFAULT_NUM_CHANNELS
from birder.model_registry import registry
from birder.net.ssl import barlow_twins
from birder.net.ssl import byol
from birder.net.ssl import capi
from birder.net.ssl import data2vec
from birder.net.ssl import data2vec2
from birder.net.ssl import dino_v1
from birder.net.ssl import dino_v2
from birder.net.ssl import franca
from birder.net.ssl import i_jepa
from birder.net.ssl import ibot
from birder.net.ssl import mmcr
from birder.net.ssl import simclr
from birder.net.ssl import sscd
from birder.net.ssl import vicreg

logging.disable(logging.CRITICAL)


# pylint: disable=too-many-public-methods
class TestNetSSL(unittest.TestCase):
    def test_barlow_twins(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v1_50", 0)
        net = barlow_twins.BarlowTwins(backbone, config={"projector_sizes": [512, 512, 512], "off_lambda": 0.005})

        # Test network
        out = net(
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
        )
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_byol(self) -> None:
        batch_size = 2
        backbone = registry.net_factory("resnet_v1_18", 0)
        net = byol.BYOL(backbone, config={"projection_size": 64, "projection_hidden_size": 128})

        # Test network
        out = net(torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_capi(self) -> None:
        batch_size = 4
        size = (192, 192)
        num_clusters = 320
        backbone = registry.net_factory("vit_s16", 0, size=size)
        input_size = (size[0] // backbone.max_stride, size[1] // backbone.max_stride)
        seq_len = input_size[0] * input_size[1]
        n_masked = int(seq_len * 0.65)
        prediction_subsampling = 0.05

        teacher = capi.CAPITeacher(
            backbone,
            config={
                "num_clusters": num_clusters,
                "bias": True,
                "n_sk_iter": 3,
                "target_temp": 0.06,
                "pred_temp": 0.12,
                "sk_mode": "position-wise",
            },
        )
        student = capi.CAPIStudent(
            backbone, config={"decoder_layers": 4, "decoder_dim": 256, "num_clusters": num_clusters}
        )
        mask_generator = masking.InverseRollBlockMasking(input_size, n_masked)

        masks = mask_generator(batch_size)
        ids_keep = masking.get_ids_keep(masks)
        n_predict = int(n_masked * prediction_subsampling)
        predict_indices = masking.get_random_masked_indices(masks, n_predict)
        masks = masking.mask_from_indices(predict_indices, seq_len)

        x = torch.rand(batch_size, DEFAULT_NUM_CHANNELS, *size)
        # all_ids = torch.arange(input_size[0] * input_size[1]).unsqueeze(0).repeat(batch_size, 1)

        #
        # Simulate a full step
        #

        # Teacher
        selected_assignments, clustering_loss = teacher(x, None, predict_indices)
        self.assertFalse(torch.isnan(clustering_loss).any())
        self.assertFalse(torch.isnan(selected_assignments).any())
        self.assertEqual(selected_assignments.size(), (batch_size * n_predict, num_clusters))

        # Student
        pred = student(x, ids_keep, predict_indices)
        self.assertEqual(pred.size(), (masks.count_nonzero().item(), num_clusters))
        self.assertFalse(torch.isnan(pred).any())

        # Loss
        loss = -torch.sum(selected_assignments * F.log_softmax(pred / 0.12, dim=-1), dim=-1)
        self.assertFalse(torch.isnan(loss.sum() / len(loss)).any())
        self.assertEqual(loss.sum().ndim, 0)

        # Test with Hiera backbone
        backbone = registry.net_factory("hiera_tiny", 0, size=size)
        input_size = (size[0] // backbone.max_stride, size[1] // backbone.max_stride)
        seq_len = input_size[0] * input_size[1]
        n_masked = int(seq_len * 0.65)
        prediction_subsampling = 0.05

        teacher = capi.CAPITeacher(
            backbone,
            config={
                "num_clusters": num_clusters,
                "bias": True,
                "n_sk_iter": 3,
                "target_temp": 0.06,
                "pred_temp": 0.12,
                "sk_mode": "position-wise",
            },
        )
        student = capi.CAPIStudent(
            backbone, config={"decoder_layers": 4, "decoder_dim": 256, "num_clusters": num_clusters}
        )
        mask_generator = masking.InverseRollBlockMasking(input_size, n_masked)

        masks = mask_generator(batch_size)
        ids_keep = masking.get_ids_keep(masks)
        n_predict = int(n_masked * prediction_subsampling)
        predict_indices = masking.get_random_masked_indices(masks, n_predict)
        masks = masking.mask_from_indices(predict_indices, seq_len)

        x = torch.rand(batch_size, DEFAULT_NUM_CHANNELS, *size)

        # Teacher
        selected_assignments, clustering_loss = teacher(x, None, predict_indices)
        self.assertFalse(torch.isnan(clustering_loss).any())
        self.assertFalse(torch.isnan(selected_assignments).any())
        self.assertEqual(selected_assignments.size(), (batch_size * n_predict, num_clusters))

        # Student
        pred = student(x, ids_keep, predict_indices)
        self.assertEqual(pred.size(), (masks.count_nonzero().item(), num_clusters))
        self.assertFalse(torch.isnan(pred).any())

        # Test with global Sinkhorn Knopp
        backbone = registry.net_factory("vit_s16", 0, size=size)
        input_size = (size[0] // backbone.max_stride, size[1] // backbone.max_stride)
        seq_len = input_size[0] * input_size[1]
        n_masked = int(seq_len * 0.65)
        prediction_subsampling = 0.05

        teacher = capi.CAPITeacher(
            backbone,
            config={
                "num_clusters": num_clusters,
                "bias": True,
                "n_sk_iter": 3,
                "target_temp": 0.06,
                "pred_temp": 0.12,
                "sk_mode": "global",
            },
        )
        student = capi.CAPIStudent(
            backbone, config={"decoder_layers": 4, "decoder_dim": 256, "num_clusters": num_clusters}
        )
        mask_generator = masking.InverseRollBlockMasking(input_size, n_masked)

        masks = mask_generator(batch_size)
        ids_keep = masking.get_ids_keep(masks)
        n_predict = int(n_masked * prediction_subsampling)
        predict_indices = masking.get_random_masked_indices(masks, n_predict)
        masks = masking.mask_from_indices(predict_indices, seq_len)

        x = torch.rand(batch_size, DEFAULT_NUM_CHANNELS, *size)

        # Teacher
        selected_assignments, clustering_loss = teacher(x, None, predict_indices)
        self.assertFalse(torch.isnan(clustering_loss).any())
        self.assertFalse(torch.isnan(selected_assignments).any())
        self.assertEqual(selected_assignments.size(), (batch_size * n_predict, num_clusters))

        # Student
        pred = student(x, ids_keep, predict_indices)
        self.assertEqual(pred.size(), (masks.count_nonzero().item(), num_clusters))
        self.assertFalse(torch.isnan(pred).any())

    def test_capi_queue(self) -> None:  # pylint: disable=too-many-locals
        batch_size = 4
        size = (192, 192)
        num_clusters = 320
        backbone = registry.net_factory("vit_s16", 0, size=size)
        input_size = (size[0] // backbone.max_stride, size[1] // backbone.max_stride)
        seq_len = input_size[0] * input_size[1]
        n_masked = int(seq_len * 0.65)
        prediction_subsampling = 0.05

        teacher = capi.CAPITeacher(
            backbone,
            config={
                "num_clusters": num_clusters,
                "bias": True,
                "n_sk_iter": 3,
                "target_temp": 0.06,
                "pred_temp": 0.12,
                "sk_mode": "position-wise",
                "queue_size": 8,
            },
        )
        student = capi.CAPIStudent(
            backbone, config={"decoder_layers": 4, "decoder_dim": 256, "num_clusters": num_clusters}
        )
        mask_generator = masking.InverseRollBlockMasking(input_size, n_masked)

        masks = mask_generator(batch_size)
        ids_keep = masking.get_ids_keep(masks)
        n_predict = int(n_masked * prediction_subsampling)
        predict_indices = masking.get_random_masked_indices(masks, n_predict)
        masks = masking.mask_from_indices(predict_indices, seq_len)

        x = torch.rand(batch_size, DEFAULT_NUM_CHANNELS, *size)

        #
        # Simulate a few full steps
        #

        teacher_logits_list = []
        for _ in range(3):
            with torch.no_grad():
                tokens = teacher.backbone.masked_encoding_omission(x, None)["tokens"]
                tokens = tokens[:, teacher.backbone.num_special_tokens :, :]
                head_in = tokens.transpose(0, 1)
                logits = teacher.head.layer(F.normalize(head_in, dim=-1, p=2, eps=1e-7))
                teacher_logits_list.append(logits.detach().clone())
                teacher.head(head_in)

        q = teacher.head.sinkhorn_queue
        assert q is not None

        # Queue state tracing:
        # - After iteration 1: positions 0-3 filled, queue_ptr = 4
        # - After iteration 2: positions 4-7 filled, queue_ptr = 0
        # - After iteration 3: positions 0-3 filled, queue_ptr = 4

        # Leftover from previous run
        self.assertTrue(torch.equal(q.queue[:, 4:8, :], teacher_logits_list[-2]))

        # Start of this run
        self.assertTrue(torch.equal(q.queue[:, :4, :], teacher_logits_list[-1]))

        # Teacher
        selected_assignments, clustering_loss = teacher(x, None, predict_indices)
        self.assertFalse(torch.isnan(clustering_loss).any())
        self.assertFalse(torch.isnan(selected_assignments).any())
        self.assertEqual(selected_assignments.size(), (batch_size * n_predict, num_clusters))

        # Student
        pred = student(x, ids_keep, predict_indices)
        self.assertEqual(pred.size(), (masks.count_nonzero().item(), num_clusters))
        self.assertFalse(torch.isnan(pred).any())

        # Loss
        loss = -torch.sum(selected_assignments * F.log_softmax(pred / 0.12, dim=-1), dim=-1)
        self.assertFalse(torch.isnan(loss.sum() / len(loss)).any())
        self.assertEqual(loss.sum().ndim, 0)

        # Test with global Sinkhorn Knopp
        backbone = registry.net_factory("vit_s16", 0, size=size)
        input_size = (size[0] // backbone.max_stride, size[1] // backbone.max_stride)
        seq_len = input_size[0] * input_size[1]
        n_masked = int(seq_len * 0.65)
        prediction_subsampling = 0.05

        teacher = capi.CAPITeacher(
            backbone,
            config={
                "num_clusters": num_clusters,
                "bias": True,
                "n_sk_iter": 3,
                "target_temp": 0.06,
                "pred_temp": 0.12,
                "sk_mode": "global",
                "queue_size": 2,
            },
        )
        student = capi.CAPIStudent(
            backbone, config={"decoder_layers": 4, "decoder_dim": 256, "num_clusters": num_clusters}
        )
        mask_generator = masking.InverseRollBlockMasking(input_size, n_masked)

        masks = mask_generator(batch_size)
        ids_keep = masking.get_ids_keep(masks)
        n_predict = int(n_masked * prediction_subsampling)
        predict_indices = masking.get_random_masked_indices(masks, n_predict)
        masks = masking.mask_from_indices(predict_indices, seq_len)

        x = torch.rand(batch_size, DEFAULT_NUM_CHANNELS, *size)

        #
        # Simulate a few full steps
        #

        teacher_logits_list = []
        for _ in range(3):
            with torch.no_grad():
                tokens = teacher.backbone.masked_encoding_omission(x, None)["tokens"]
                tokens = tokens[:, teacher.backbone.num_special_tokens :, :]
                head_in = tokens.transpose(0, 1)
                logits = teacher.head.layer(F.normalize(head_in, dim=-1, p=2, eps=1e-7))
                teacher_logits_list.append(logits.detach().clone())
                teacher.head(head_in)

        q = teacher.head.sinkhorn_queue
        assert q is not None
        logits_flat = teacher_logits_list[-1].flatten(0, -2)
        self.assertTrue(torch.equal(q.queue, logits_flat[-q.queue_size :]))

        # Teacher
        selected_assignments, clustering_loss = teacher(x, None, predict_indices)
        self.assertFalse(torch.isnan(clustering_loss).any())
        self.assertFalse(torch.isnan(selected_assignments).any())
        self.assertEqual(selected_assignments.size(), (batch_size * n_predict, num_clusters))

        # Student
        pred = student(x, ids_keep, predict_indices)
        self.assertEqual(pred.size(), (masks.count_nonzero().item(), num_clusters))
        self.assertFalse(torch.isnan(pred).any())

        # Loss
        loss = -torch.sum(selected_assignments * F.log_softmax(pred / 0.12, dim=-1), dim=-1)
        self.assertFalse(torch.isnan(loss.sum() / len(loss)).any())
        self.assertEqual(loss.sum().ndim, 0)

    def test_capi_sk_no_mutation(self) -> None:
        # Global with queue
        head = capi.OnlineClustering(
            16, 32, bias=True, n_sk_iter=3, target_temp=0.06, pred_temp=0.12, position_wise_sk=False, queue_size=4
        )
        for _ in range(3):
            x = torch.randn(4, 8, 16)
            x_clone = x.clone()
            head(x)
            self.assertTrue(torch.equal(x, x_clone))

        # Global without queue
        head = capi.OnlineClustering(
            16, 32, bias=True, n_sk_iter=3, target_temp=0.06, pred_temp=0.12, position_wise_sk=False, queue_size=None
        )
        for _ in range(3):
            x = torch.randn(4, 8, 16)
            x_clone = x.clone()
            head(x)
            self.assertTrue(torch.equal(x, x_clone))

        # Position-wise with queue
        head = capi.OnlineClustering(
            16,
            32,
            bias=True,
            n_sk_iter=3,
            target_temp=0.06,
            pred_temp=0.12,
            position_wise_sk=True,
            queue_size=4,
            seq_len=4,
        )
        for _ in range(3):
            x = torch.randn(4, 8, 16)
            x_clone = x.clone()
            head(x)
            self.assertTrue(torch.equal(x, x_clone))

        # Position-wise without queue
        head = capi.OnlineClustering(
            16, 32, bias=True, n_sk_iter=3, target_temp=0.06, pred_temp=0.12, position_wise_sk=True, queue_size=None
        )
        for _ in range(3):
            x = torch.randn(4, 8, 16)
            x_clone = x.clone()
            head(x)
            self.assertTrue(torch.equal(x, x_clone))

    def test_data2vec(self) -> None:
        batch_size = 2
        backbone = registry.net_factory("vit_t16", 0, size=(96, 96))
        net = data2vec.Data2Vec(
            backbone, config={"normalize_targets": True, "average_top_k_layers": 6, "loss_beta": 2.0}
        )

        mask_generator = masking.BlockMasking((96 // backbone.stem_stride, 96 // backbone.stem_stride), 1, 3, 0.66, 1.5)
        masks = mask_generator(batch_size)

        # Test network
        out = net(torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)), masks)
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_data2vec2(self) -> None:
        batch_size = 2
        backbone = registry.net_factory("vit_t16", 0, size=(128, 128))
        net = data2vec2.Data2Vec2(
            backbone,
            config={
                "average_top_k_layers": 6,
                "decoder_dim": 128,
                "decoder_kernel_size": 3,
                "decoder_layers": 4,
                "clone_batch": 3,
                "cls_loss_weight": 0.1,
            },
        )

        mask_generator = masking.InverseRollBlockMasking(
            (128 // backbone.stem_stride, 128 // backbone.stem_stride),
            num_masking_patches=32,
            min_aspect=0.66,
            max_aspect=1.5,
        )
        masks = mask_generator(batch_size * 3)

        # Test network
        out = net(torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)), masks)
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_dino_v1(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v2_18", 0)
        net = dino_v1.DINO_v1(
            backbone,
            config={
                "out_dim": 128,
                "use_bn": True,
                "norm_last_layer": True,
                "num_layers": 3,
                "hidden_dim": 512,
                "bottleneck_dim": 256,
            },
        )

        # Test network
        out = net(
            [
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
            ]
        )
        self.assertFalse(torch.isnan(out).any())

        teacher_out = net(
            [
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
                torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
            ]
        )
        dino_loss = dino_v1.DINOLoss(
            128,
            6,
            warmup_teacher_temp=0.2,
            teacher_temp=0.8,
            warmup_teacher_temp_epochs=10,
            num_epochs=100,
            student_temp=0.5,
            center_momentum=0.99,
        )
        loss = dino_loss(out, teacher_out, epoch=2)
        self.assertFalse(torch.isnan(loss).any())
        self.assertEqual(loss.ndim, 0)

    # pylint: disable=too-many-locals
    def test_dino_v2(self) -> None:
        batch_size = 4
        size = (192, 192)
        backbone = registry.net_factory("vit_s16", 0, size=size)
        backbone.set_dynamic_size()

        # Without iBOT head
        teacher = dino_v2.DINOv2Teacher(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
            },
        )
        student = dino_v2.DINOv2Student(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
            },
        )

        x = torch.rand(batch_size * 2, DEFAULT_NUM_CHANNELS, *size)
        local_x = torch.rand(batch_size * 4, DEFAULT_NUM_CHANNELS, size[0] // 2, size[1] // 2)
        mask_generator = masking.BlockMasking(
            (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)  # Not quite the same, but close enough
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16

        with torch.no_grad():
            teacher_embedding_after_head, teacher_masked_patch_tokens_after_head = teacher(
                x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
            )

        (
            student_global_embedding,
            student_global_embedding_after_head,
            student_local_embedding_after_head,
            student_global_masked_patch_tokens_after_head,
        ) = student(x, local_x, masks, upper_bound, mask_indices_list)

        self.assertFalse(torch.isnan(teacher_embedding_after_head).any())
        self.assertEqual(teacher_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(teacher_masked_patch_tokens_after_head).any())
        self.assertEqual(teacher_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        self.assertFalse(torch.isnan(student_global_embedding).any())
        self.assertEqual(student_global_embedding.size(), (batch_size * 2, backbone.embedding_size))
        self.assertFalse(torch.isnan(student_global_embedding_after_head).any())
        self.assertEqual(student_global_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(student_local_embedding_after_head).any())
        self.assertEqual(student_local_embedding_after_head.size(), (batch_size * 4, 4096))
        self.assertFalse(torch.isnan(student_global_masked_patch_tokens_after_head).any())
        self.assertEqual(student_global_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        # With iBOT head
        teacher = dino_v2.DINOv2Teacher(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": True,
                "ibot_out_dim": 4096,
            },
        )
        student = dino_v2.DINOv2Student(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": True,
                "ibot_out_dim": 4096,
            },
        )

        x = torch.rand(batch_size * 2, DEFAULT_NUM_CHANNELS, *size)
        local_x = torch.rand(batch_size * 4, DEFAULT_NUM_CHANNELS, size[0] // 2, size[1] // 2)
        mask_generator = masking.BlockMasking(
            (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)  # Not quite the same, but close enough
        masks = masks.to(torch.long)
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16

        with torch.no_grad():
            teacher_embedding_after_head, teacher_masked_patch_tokens_after_head = teacher(
                x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
            )

        (
            student_global_embedding,
            student_global_embedding_after_head,
            student_local_embedding_after_head,
            student_global_masked_patch_tokens_after_head,
        ) = student(x, local_x, masks, upper_bound, mask_indices_list)

        self.assertFalse(torch.isnan(teacher_embedding_after_head).any())
        self.assertEqual(teacher_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(teacher_masked_patch_tokens_after_head).any())
        self.assertEqual(teacher_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        self.assertFalse(torch.isnan(student_global_embedding).any())
        self.assertEqual(student_global_embedding.size(), (batch_size * 2, backbone.embedding_size))
        self.assertFalse(torch.isnan(student_global_embedding_after_head).any())
        self.assertEqual(student_global_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(student_local_embedding_after_head).any())
        self.assertEqual(student_local_embedding_after_head.size(), (batch_size * 4, 4096))
        self.assertFalse(torch.isnan(student_global_masked_patch_tokens_after_head).any())
        self.assertEqual(student_global_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        # Loss
        dino_loss = dino_v2.DINOLoss(4096, student_temp=0.1, center_momentum=0.9)
        koleo_loss = dino_v2.KoLeoLoss()
        ibot_patch_loss = dino_v2.iBOTPatchLoss(4096, student_temp=0.1, center_momentum=0.9)

        # Loss centering
        teacher_temp = 0.04
        n_masked_patches = len(mask_indices_list)

        teacher_dino_softmax_centered = dino_loss.softmax_center_teacher(
            teacher_embedding_after_head, teacher_temp=teacher_temp
        ).view(2, -1, *teacher_embedding_after_head.shape[1:])
        dino_loss.update_center(teacher_embedding_after_head)

        teacher_masked_patch_tokens_after_head_d = teacher_masked_patch_tokens_after_head.unsqueeze(0)
        masked_teacher_ibot_softmax_centered = ibot_patch_loss.softmax_center_teacher(
            teacher_masked_patch_tokens_after_head_d[:, :n_masked_patches], teacher_temp=teacher_temp
        )
        masked_teacher_ibot_softmax_centered = masked_teacher_ibot_softmax_centered.squeeze(0)
        ibot_patch_loss.update_center(teacher_masked_patch_tokens_after_head_d[:n_masked_patches])

        self.assertFalse(torch.isnan(teacher_dino_softmax_centered).any())
        self.assertEqual(teacher_dino_softmax_centered.size(), (2, batch_size, 4096))
        self.assertFalse(torch.isnan(masked_teacher_ibot_softmax_centered).any())
        self.assertEqual(masked_teacher_ibot_softmax_centered.size(), (len(mask_indices_list), 4096))

        # Loss centering - Sinkhorn Knopp
        teacher_dino_softmax_centered = dino_loss.sinkhorn_knopp_teacher(
            teacher_embedding_after_head, teacher_temp=teacher_temp
        ).view(2, -1, *teacher_embedding_after_head.shape[1:])

        masked_teacher_ibot_softmax_centered = ibot_patch_loss.sinkhorn_knopp_teacher(
            teacher_masked_patch_tokens_after_head,
            teacher_temp=teacher_temp,
            n_masked_patches_tensor=torch.full((1,), fill_value=mask_indices_list.size(0), dtype=torch.long),
        )

        self.assertFalse(torch.isnan(teacher_dino_softmax_centered).any())
        self.assertEqual(teacher_dino_softmax_centered.size(), (2, batch_size, 4096))
        self.assertFalse(torch.isnan(masked_teacher_ibot_softmax_centered).any())
        self.assertEqual(masked_teacher_ibot_softmax_centered.size(), (len(mask_indices_list), 4096))

        # DINO loss
        loss_dino = dino_loss(student_local_embedding_after_head.chunk(4), teacher_dino_softmax_centered.unbind(0))
        self.assertFalse(torch.isnan(loss_dino).any())
        self.assertEqual(loss_dino.ndim, 0)

        # KoLeo loss
        loss_koleo = sum(koleo_loss(p) for p in student_global_embedding.chunk(2))
        self.assertFalse(torch.isnan(loss_koleo).any())
        self.assertEqual(loss_koleo.ndim, 0)

        # iBOT loss
        masks_weight = (1 / masks.sum(-1).clamp(min=1.0)).unsqueeze(-1).expand_as(masks)[masks.bool()]
        loss_ibot_patch = ibot_patch_loss(
            student_global_masked_patch_tokens_after_head,
            masked_teacher_ibot_softmax_centered,
            student_masks_flat=masks,
            masks_weight=masks_weight,
            n_masked_patches=n_masked_patches,
        )
        self.assertFalse(torch.isnan(loss_ibot_patch).any())
        self.assertEqual(loss_ibot_patch.ndim, 0)

    def test_dino_v2_loss_forward_matches_reference(self) -> None:
        torch.manual_seed(0)
        batch_size = 3
        out_dim = 16
        n_students = 4
        n_teachers = 2

        dino_loss = dino_v2.DINOLoss(out_dim, student_temp=0.1, center_momentum=0.9)

        # Student outputs are logits (any real values)
        student_output_list = [torch.randn(batch_size, out_dim, requires_grad=True) for _ in range(n_students)]

        # Teacher outputs should be valid probability distributions (softmax outputs)
        teacher_out_softmax_centered_list = [
            F.softmax(torch.randn(batch_size, out_dim), dim=-1) for _ in range(n_teachers)
        ]

        # Test forward pass
        loss_reference = dino_loss.forward_reference(student_output_list, teacher_out_softmax_centered_list)
        loss_vectorized = dino_loss.forward(student_output_list, teacher_out_softmax_centered_list)

        self.assertEqual(loss_reference.ndim, 0)
        self.assertEqual(loss_vectorized.ndim, 0)
        self.assertTrue(torch.allclose(loss_reference, loss_vectorized, rtol=1e-5, atol=1e-5))

        # Test gradients match
        loss_reference.backward()
        grads_reference = [s.grad.clone() for s in student_output_list]

        for s in student_output_list:
            s.grad = None

        loss_vectorized = dino_loss.forward(student_output_list, teacher_out_softmax_centered_list)
        loss_vectorized.backward()

        for g_reference, s in zip(grads_reference, student_output_list):
            self.assertTrue(torch.allclose(g_reference, s.grad, rtol=1e-5, atol=1e-5))

    def test_dino_v2_queue(self) -> None:
        batch_size = 4
        size = (192, 192)
        backbone = registry.net_factory("vit_s16", 0, size=size)
        backbone.set_dynamic_size()

        # Without iBOT head
        teacher = dino_v2.DINOv2Teacher(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
            },
        )
        student = dino_v2.DINOv2Student(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
            },
        )

        x = torch.rand(batch_size * 2, DEFAULT_NUM_CHANNELS, *size)
        local_x = torch.rand(batch_size * 4, DEFAULT_NUM_CHANNELS, size[0] // 2, size[1] // 2)
        mask_generator = masking.BlockMasking(
            (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)  # Not quite the same, but close enough
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16
        teacher_temp = 0.04
        n_masked_patches = len(mask_indices_list)

        # Loss
        dino_loss = dino_v2.DINOLoss(4096, student_temp=0.1, center_momentum=0.9, queue_size=10)
        koleo_loss = dino_v2.KoLeoLoss()
        ibot_patch_loss = dino_v2.iBOTPatchLoss(4096, student_temp=0.1, center_momentum=0.9, queue_size=10)

        teacher_embedding_after_head_list = []
        for _ in range(3):
            with torch.no_grad():
                teacher_embedding_after_head, teacher_masked_patch_tokens_after_head = teacher(
                    x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
                )

            (
                student_global_embedding,
                _student_global_embedding_after_head,
                student_local_embedding_after_head,
                student_global_masked_patch_tokens_after_head,
            ) = student(x, local_x, masks, upper_bound, mask_indices_list)

            teacher_embedding_after_head_list.append(teacher_embedding_after_head)

            teacher_dino_softmax_centered = dino_loss.sinkhorn_knopp_teacher(
                teacher_embedding_after_head, teacher_temp=teacher_temp
            ).view(2, -1, *teacher_embedding_after_head.shape[1:])

            masked_teacher_ibot_softmax_centered = ibot_patch_loss.sinkhorn_knopp_teacher(
                teacher_masked_patch_tokens_after_head,
                teacher_temp=teacher_temp,
                n_masked_patches_tensor=torch.full((1,), fill_value=mask_indices_list.size(0), dtype=torch.long),
            )

            self.assertFalse(torch.isnan(teacher_dino_softmax_centered).any())
            self.assertEqual(teacher_dino_softmax_centered.size(), (2, batch_size, 4096))
            self.assertFalse(torch.isnan(masked_teacher_ibot_softmax_centered).any())
            self.assertEqual(masked_teacher_ibot_softmax_centered.size(), (len(mask_indices_list), 4096))

            loss_dino = dino_loss(student_local_embedding_after_head.chunk(4), teacher_dino_softmax_centered.unbind(0))
            loss_koleo = sum(koleo_loss(p) for p in student_global_embedding.chunk(2))
            masks_weight = (1 / masks.sum(-1).clamp(min=1.0)).unsqueeze(-1).expand_as(masks)[masks.bool()]
            loss_ibot_patch = ibot_patch_loss(
                student_global_masked_patch_tokens_after_head,
                masked_teacher_ibot_softmax_centered,
                student_masks_flat=masks,
                masks_weight=masks_weight,
                n_masked_patches=n_masked_patches,
            )

            # DINO loss
            self.assertFalse(torch.isnan(loss_dino).any())
            self.assertEqual(loss_dino.ndim, 0)

            # KoLeo loss
            self.assertFalse(torch.isnan(loss_koleo).any())
            self.assertEqual(loss_koleo.ndim, 0)

            # iBOT loss
            self.assertFalse(torch.isnan(loss_ibot_patch).any())
            self.assertEqual(loss_ibot_patch.ndim, 0)

        q = dino_loss.sinkhorn_queue
        assert q is not None

        # Queue state tracing:
        # - After iteration 1: positions 0-7 filled, queue_ptr = 8
        # - After iteration 2: write at 8-9, wrap to 0-5, queue_ptr = 6
        # - After iteration 3: write at 6-9, wrap to 0-3, queue_ptr = 4

        # Leftover from previous run
        self.assertTrue(
            torch.equal(
                q.queue[4:6],  # queue_ptr : queue_size - queue_ptr
                teacher_embedding_after_head_list[-2][-2:],
            )
        )

        # Start of this run
        self.assertTrue(
            torch.equal(
                q.queue[6:10],  # queue_size - queue_ptr : queue_size
                teacher_embedding_after_head_list[-1][:4],
            )
        )

        # Rollover of this run
        self.assertTrue(
            torch.equal(
                q.queue[:4],  # 0 : queue_ptr
                teacher_embedding_after_head_list[-1][4:],
            )
        )

    def test_dino_v2_sk_no_mutation(self) -> None:
        dino_loss = dino_v2.DINOLoss(256, student_temp=0.1, center_momentum=0.9, queue_size=2)
        ibot_patch_loss = dino_v2.iBOTPatchLoss(256, student_temp=0.1, center_momentum=0.9, queue_size=2)

        # Run a few steps to fill the queue
        for _ in range(3):
            teacher_output = torch.randn(8, 256)
            teacher_output_clone = teacher_output.clone()
            dino_loss.sinkhorn_knopp_teacher(teacher_output, teacher_temp=0.04)
            self.assertTrue(torch.equal(teacher_output, teacher_output_clone))

            patch_output = torch.randn(16, 256)
            patch_output_clone = patch_output.clone()
            ibot_patch_loss.sinkhorn_knopp_teacher(
                patch_output, teacher_temp=0.04, n_masked_patches_tensor=torch.tensor([16])
            )
            self.assertTrue(torch.equal(patch_output, patch_output_clone))

        dino_loss = dino_v2.DINOLoss(256, student_temp=0.1, center_momentum=0.9, queue_size=None)
        ibot_patch_loss = dino_v2.iBOTPatchLoss(256, student_temp=0.1, center_momentum=0.9, queue_size=None)

        for _ in range(3):
            teacher_output = torch.randn(8, 256)
            teacher_output_clone = teacher_output.clone()
            dino_loss.sinkhorn_knopp_teacher(teacher_output, teacher_temp=0.04)
            self.assertTrue(torch.equal(teacher_output, teacher_output_clone))

            patch_output = torch.randn(16, 256)
            patch_output_clone = patch_output.clone()
            ibot_patch_loss.sinkhorn_knopp_teacher(
                patch_output, teacher_temp=0.04, n_masked_patches_tensor=torch.tensor([16])
            )
            self.assertTrue(torch.equal(patch_output, patch_output_clone))

    # pylint: disable=too-many-locals
    def test_franca(self) -> None:
        batch_size = 4
        size = (192, 192)
        dino_out_dim = 4096
        num_nesting_levels = 3
        backbone = registry.net_factory("vit_s16", 0, size=size)
        backbone.set_dynamic_size()

        # Without iBOT head
        teacher = franca.FrancaTeacher(
            backbone,
            config={
                "dino_out_dim": dino_out_dim,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
                "nesting_levels": num_nesting_levels,
            },
        )
        student = franca.FrancaStudent(
            backbone,
            config={
                "dino_out_dim": dino_out_dim,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
                "nesting_levels": num_nesting_levels,
            },
        )

        x = torch.rand(batch_size * 2, DEFAULT_NUM_CHANNELS, *size)
        local_x = torch.rand(batch_size * 4, DEFAULT_NUM_CHANNELS, size[0] // 2, size[1] // 2)
        mask_generator = masking.BlockMasking(
            (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)  # Not quite the same, but close enough
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16

        with torch.no_grad():
            teacher_embedding_after_head, teacher_masked_patch_tokens_after_head = teacher(
                x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
            )

        (
            student_global_embedding,
            student_global_embedding_after_head,
            student_local_embedding_after_head,
            student_global_masked_patch_tokens_after_head,
        ) = student(x, local_x, masks, upper_bound, mask_indices_list)

        # Teacher outputs are tuples (one per nesting level)
        self.assertEqual(len(teacher_embedding_after_head), num_nesting_levels)
        self.assertEqual(len(teacher_masked_patch_tokens_after_head), num_nesting_levels)
        for i, t in enumerate(teacher_embedding_after_head):
            expected_dim = dino_out_dim // (2 ** (num_nesting_levels - 1 - i))
            self.assertFalse(torch.isnan(t).any())
            self.assertEqual(t.size(), (batch_size * 2, expected_dim))

        for i, t in enumerate(teacher_masked_patch_tokens_after_head):
            expected_dim = dino_out_dim // (2 ** (num_nesting_levels - 1 - i))
            self.assertFalse(torch.isnan(t).any())
            self.assertEqual(t.size(), (len(mask_indices_list), expected_dim))

        # Student outputs
        self.assertFalse(torch.isnan(student_global_embedding).any())
        self.assertEqual(student_global_embedding.size(), (batch_size * 2, backbone.embedding_size))

        self.assertEqual(len(student_global_embedding_after_head), num_nesting_levels)
        self.assertEqual(len(student_local_embedding_after_head), num_nesting_levels)
        self.assertEqual(len(student_global_masked_patch_tokens_after_head), num_nesting_levels)

        for i, t in enumerate(student_global_embedding_after_head):
            expected_dim = dino_out_dim // (2 ** (num_nesting_levels - 1 - i))
            self.assertFalse(torch.isnan(t).any())
            self.assertEqual(t.size(), (batch_size * 2, expected_dim))

        for i, t in enumerate(student_local_embedding_after_head):
            expected_dim = dino_out_dim // (2 ** (num_nesting_levels - 1 - i))
            self.assertFalse(torch.isnan(t).any())
            self.assertEqual(t.size(), (batch_size * 4, expected_dim))

        for i, t in enumerate(student_global_masked_patch_tokens_after_head):
            expected_dim = dino_out_dim // (2 ** (num_nesting_levels - 1 - i))
            self.assertFalse(torch.isnan(t).any())
            self.assertEqual(t.size(), (len(mask_indices_list), expected_dim))

        # With iBOT head
        teacher = franca.FrancaTeacher(
            backbone,
            config={
                "dino_out_dim": dino_out_dim,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": True,
                "ibot_out_dim": dino_out_dim,
                "nesting_levels": num_nesting_levels,
            },
        )
        student = franca.FrancaStudent(
            backbone,
            config={
                "dino_out_dim": dino_out_dim,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": True,
                "ibot_out_dim": dino_out_dim,
                "nesting_levels": num_nesting_levels,
            },
        )

        masks = mask_generator(batch_size * 2)
        masks = masks.to(torch.long)
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16

        with torch.no_grad():
            teacher_embedding_after_head, teacher_masked_patch_tokens_after_head = teacher(
                x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
            )

        (
            student_global_embedding,
            student_global_embedding_after_head,
            student_local_embedding_after_head,
            student_global_masked_patch_tokens_after_head,
        ) = student(x, local_x, masks, upper_bound, mask_indices_list)

        # Verify shapes again with separate iBOT head
        self.assertEqual(len(teacher_embedding_after_head), num_nesting_levels)
        self.assertEqual(len(teacher_masked_patch_tokens_after_head), num_nesting_levels)

        # Loss - Sinkhorn-Knopp only (centering not supported for MRL)
        dino_loss = franca.DINOLossMRL(student_temp=0.1, nesting_levels=num_nesting_levels)
        koleo_loss = dino_v2.KoLeoLoss()
        ibot_patch_loss = franca.iBOTPatchLossMRL(student_temp=0.1, nesting_levels=num_nesting_levels)

        teacher_temp = 0.04
        n_masked_patches = len(mask_indices_list)

        # Sinkhorn-Knopp centering
        teacher_dino_softmax_centered_list = dino_loss.sinkhorn_knopp_teacher(
            teacher_embedding_after_head, teacher_temp=teacher_temp
        )

        masked_teacher_ibot_softmax_centered = ibot_patch_loss.sinkhorn_knopp_teacher(
            teacher_masked_patch_tokens_after_head,
            teacher_temp=teacher_temp,
            n_masked_patches_tensor=torch.full((1,), fill_value=mask_indices_list.size(0), dtype=torch.long),
        )

        self.assertEqual(len(teacher_dino_softmax_centered_list), num_nesting_levels)
        self.assertEqual(len(masked_teacher_ibot_softmax_centered), num_nesting_levels)
        for t in teacher_dino_softmax_centered_list:
            self.assertFalse(torch.isnan(t).any())
        for t in masked_teacher_ibot_softmax_centered:
            self.assertFalse(torch.isnan(t).any())

        # DINO loss
        n_local_crops = 4
        n_global_crops = 2
        loss_dino = dino_loss(
            list(student_local_embedding_after_head),
            teacher_dino_softmax_centered_list,
            n_crops=(n_local_crops, n_global_crops),
            teacher_global=False,
        )
        self.assertFalse(torch.isnan(loss_dino).any())
        self.assertEqual(loss_dino.ndim, 0)

        # KoLeo loss
        loss_koleo = sum(koleo_loss(p) for p in student_global_embedding.chunk(2))
        self.assertFalse(torch.isnan(loss_koleo).any())
        self.assertEqual(loss_koleo.ndim, 0)

        # iBOT loss
        masks_weight = (1 / masks.sum(-1).clamp(min=1.0)).unsqueeze(-1).expand_as(masks)[masks.bool()]
        loss_ibot_patch = ibot_patch_loss(
            student_global_masked_patch_tokens_after_head,
            masked_teacher_ibot_softmax_centered,
            student_masks_flat=masks,
            masks_weight=masks_weight,
            n_masked_patches=n_masked_patches,
        )
        self.assertFalse(torch.isnan(loss_ibot_patch).any())
        self.assertEqual(loss_ibot_patch.ndim, 0)

    def test_franca_loss_forward_matches_reference(self) -> None:
        torch.manual_seed(0)
        batch_size = 3
        out_dim = 16
        n_local_crops = 4
        n_global_crops = 2
        num_nesting_levels = 3

        dino_loss = franca.DINOLossMRL(student_temp=0.1, nesting_levels=num_nesting_levels)

        # Student outputs are logits (any real values) - one tensor per nesting level
        student_output_list = [
            torch.randn(batch_size * n_local_crops, out_dim // (2 ** (num_nesting_levels - 1 - i)), requires_grad=True)
            for i in range(num_nesting_levels)
        ]

        # Teacher outputs should be valid probability distributions - one tensor per nesting level
        teacher_out_softmax_centered_list = [
            F.softmax(torch.randn(n_global_crops * batch_size, out_dim // (2 ** (num_nesting_levels - 1 - i))), dim=-1)
            for i in range(num_nesting_levels)
        ]

        # Test forward pass (teacher_global=False)
        loss_reference = dino_loss.forward_reference(
            student_output_list,
            teacher_out_softmax_centered_list,
            n_crops=(n_local_crops, n_global_crops),
            teacher_global=False,
        )
        loss_vectorized = dino_loss.forward(
            student_output_list,
            teacher_out_softmax_centered_list,
            n_crops=(n_local_crops, n_global_crops),
            teacher_global=False,
        )

        self.assertEqual(loss_reference.ndim, 0)
        self.assertEqual(loss_vectorized.ndim, 0)
        self.assertTrue(torch.allclose(loss_reference, loss_vectorized, rtol=1e-5, atol=1e-5))

        # Test gradients match (teacher_global=False)
        loss_reference.backward()
        grads_reference = [s.grad.clone() for s in student_output_list]

        for s in student_output_list:
            s.grad = None

        loss_vectorized = dino_loss.forward(
            student_output_list,
            teacher_out_softmax_centered_list,
            n_crops=(n_local_crops, n_global_crops),
            teacher_global=False,
        )
        loss_vectorized.backward()

        for g_reference, s in zip(grads_reference, student_output_list):
            self.assertTrue(torch.allclose(g_reference, s.grad, rtol=1e-5, atol=1e-5))

        # Test forward pass (teacher_global=True)
        for s in student_output_list:
            s.grad = None

        student_global_list = [
            torch.randn(batch_size * n_global_crops, out_dim // (2 ** (num_nesting_levels - 1 - i)), requires_grad=True)
            for i in range(num_nesting_levels)
        ]

        loss_reference_global = dino_loss.forward_reference(
            student_global_list, teacher_out_softmax_centered_list, n_crops=n_global_crops, teacher_global=True
        )
        loss_vectorized_global = dino_loss.forward(
            student_global_list, teacher_out_softmax_centered_list, n_crops=n_global_crops, teacher_global=True
        )

        self.assertEqual(loss_reference_global.ndim, 0)
        self.assertEqual(loss_vectorized_global.ndim, 0)
        self.assertTrue(torch.allclose(loss_reference_global, loss_vectorized_global, rtol=1e-5, atol=1e-5))

    def test_franca_queue(self) -> None:
        batch_size = 4
        size = (192, 192)
        dino_out_dim = 4096
        num_nesting_levels = 3
        backbone = registry.net_factory("vit_s16", 0, size=size)
        backbone.set_dynamic_size()

        # Without iBOT head
        teacher = franca.FrancaTeacher(
            backbone,
            config={
                "dino_out_dim": dino_out_dim,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
                "nesting_levels": num_nesting_levels,
            },
        )
        student = franca.FrancaStudent(
            backbone,
            config={
                "dino_out_dim": dino_out_dim,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
                "nesting_levels": num_nesting_levels,
            },
        )

        x = torch.rand(batch_size * 2, DEFAULT_NUM_CHANNELS, *size)
        local_x = torch.rand(batch_size * 4, DEFAULT_NUM_CHANNELS, size[0] // 2, size[1] // 2)
        mask_generator = masking.BlockMasking(
            (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)  # Not quite the same, but close enough
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16
        teacher_temp = 0.04
        n_masked_patches = len(mask_indices_list)
        n_local_crops = 4
        n_global_crops = 2

        # Loss
        dino_loss = franca.DINOLossMRL(
            student_temp=0.1,
            nesting_levels=num_nesting_levels,
            queue_size=10,
            out_dim=dino_out_dim,
        )
        koleo_loss = dino_v2.KoLeoLoss()
        ibot_patch_loss = franca.iBOTPatchLossMRL(
            student_temp=0.1,
            nesting_levels=num_nesting_levels,
            queue_size=10,
            out_dim=dino_out_dim,
        )

        teacher_embedding_after_head_list = []
        for _ in range(3):
            with torch.no_grad():
                teacher_embedding_after_head, teacher_masked_patch_tokens_after_head = teacher(
                    x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
                )

            (
                student_global_embedding,
                _student_global_embedding_after_head,
                student_local_embedding_after_head,
                student_global_masked_patch_tokens_after_head,
            ) = student(x, local_x, masks, upper_bound, mask_indices_list)

            teacher_embedding_after_head_list.append(teacher_embedding_after_head)

            teacher_dino_softmax_centered_list = dino_loss.sinkhorn_knopp_teacher(
                teacher_embedding_after_head, teacher_temp=teacher_temp
            )

            masked_teacher_ibot_softmax_centered = ibot_patch_loss.sinkhorn_knopp_teacher(
                teacher_masked_patch_tokens_after_head,
                teacher_temp=teacher_temp,
                n_masked_patches_tensor=torch.full((1,), fill_value=mask_indices_list.size(0), dtype=torch.long),
            )

            self.assertEqual(len(teacher_dino_softmax_centered_list), num_nesting_levels)
            self.assertEqual(len(masked_teacher_ibot_softmax_centered), num_nesting_levels)
            for t in teacher_dino_softmax_centered_list:
                self.assertFalse(torch.isnan(t).any())
            for t in masked_teacher_ibot_softmax_centered:
                self.assertFalse(torch.isnan(t).any())

            loss_dino = dino_loss(
                list(student_local_embedding_after_head),
                teacher_dino_softmax_centered_list,
                n_crops=(n_local_crops, n_global_crops),
                teacher_global=False,
            )
            loss_koleo = sum(koleo_loss(p) for p in student_global_embedding.chunk(2))
            masks_weight = (1 / masks.sum(-1).clamp(min=1.0)).unsqueeze(-1).expand_as(masks)[masks.bool()]
            loss_ibot_patch = ibot_patch_loss(
                student_global_masked_patch_tokens_after_head,
                masked_teacher_ibot_softmax_centered,
                student_masks_flat=masks,
                masks_weight=masks_weight,
                n_masked_patches=n_masked_patches,
            )

            # DINO loss
            self.assertFalse(torch.isnan(loss_dino).any())
            self.assertEqual(loss_dino.ndim, 0)

            # KoLeo loss
            self.assertFalse(torch.isnan(loss_koleo).any())
            self.assertEqual(loss_koleo.ndim, 0)

            # iBOT loss
            self.assertFalse(torch.isnan(loss_ibot_patch).any())
            self.assertEqual(loss_ibot_patch.ndim, 0)

        assert dino_loss.sinkhorn_queue is not None
        q0 = dino_loss.sinkhorn_queue[0]
        q1 = dino_loss.sinkhorn_queue[1]
        assert q0 is not None and q1 is not None

        # Queue state tracing:
        # - After iteration 1: positions 0-7 filled, queue_ptr = 8
        # - After iteration 2: write at 8-9, wrap to 0-5, queue_ptr = 6
        # - After iteration 3: write at 6-9, wrap to 0-3, queue_ptr = 4

        # Leftover from previous run
        self.assertTrue(
            torch.equal(
                q0.queue[4:6],  # queue_ptr : queue_size - queue_ptr
                teacher_embedding_after_head_list[-2][0][-2:],
            )
        )
        self.assertTrue(
            torch.equal(
                q1.queue[4:6],  # queue_ptr : queue_size - queue_ptr
                teacher_embedding_after_head_list[-2][1][-2:],
            )
        )

        # Start of this run
        self.assertTrue(
            torch.equal(
                q0.queue[6:10],  # queue_size - queue_ptr : queue_size
                teacher_embedding_after_head_list[-1][0][:4],
            )
        )
        self.assertTrue(
            torch.equal(
                q1.queue[6:10],  # queue_size - queue_ptr : queue_size
                teacher_embedding_after_head_list[-1][1][:4],
            )
        )

        # Rollover of this run
        self.assertTrue(
            torch.equal(
                q0.queue[:4],  # 0 : queue_ptr
                teacher_embedding_after_head_list[-1][0][4:],
            )
        )
        self.assertTrue(
            torch.equal(
                q1.queue[:4],  # 0 : queue_ptr
                teacher_embedding_after_head_list[-1][1][4:],
            )
        )

    def test_franca_sk_no_mutation(self) -> None:
        num_nesting_levels = 2
        out_dim = 256
        dino_loss = franca.DINOLossMRL(
            student_temp=0.1,
            nesting_levels=num_nesting_levels,
            queue_size=10,
            out_dim=out_dim,
        )
        ibot_patch_loss = franca.iBOTPatchLossMRL(
            student_temp=0.1,
            nesting_levels=num_nesting_levels,
            queue_size=10,
            out_dim=out_dim,
        )

        # Run a few steps to fill the queue
        for _ in range(3):
            teacher_output = (torch.randn(8, 128), torch.randn(8, 256))
            teacher_output_clone = (teacher_output[0].clone(), teacher_output[1].clone())
            dino_loss.sinkhorn_knopp_teacher(teacher_output, teacher_temp=0.04)
            self.assertTrue(torch.equal(teacher_output[0], teacher_output_clone[0]))
            self.assertTrue(torch.equal(teacher_output[1], teacher_output_clone[1]))

            patch_output = (torch.randn(16, 128), torch.randn(16, 256))
            patch_output_clone = (patch_output[0].clone(), patch_output[1].clone())
            ibot_patch_loss.sinkhorn_knopp_teacher(
                patch_output, teacher_temp=0.04, n_masked_patches_tensor=torch.tensor([16])
            )
            self.assertTrue(torch.equal(patch_output[0], patch_output_clone[0]))
            self.assertTrue(torch.equal(patch_output[1], patch_output_clone[1]))

        dino_loss = franca.DINOLossMRL(student_temp=0.1, nesting_levels=num_nesting_levels, queue_size=None)
        ibot_patch_loss = franca.iBOTPatchLossMRL(student_temp=0.1, nesting_levels=num_nesting_levels, queue_size=None)

        for _ in range(3):
            teacher_output = (torch.randn(8, 128), torch.randn(8, 256))
            teacher_output_clone = (teacher_output[0].clone(), teacher_output[1].clone())
            dino_loss.sinkhorn_knopp_teacher(teacher_output, teacher_temp=0.04)
            self.assertTrue(torch.equal(teacher_output[0], teacher_output_clone[0]))
            self.assertTrue(torch.equal(teacher_output[1], teacher_output_clone[1]))

            patch_output = (torch.randn(16, 128), torch.randn(16, 256))
            patch_output_clone = (patch_output[0].clone(), patch_output[1].clone())
            ibot_patch_loss.sinkhorn_knopp_teacher(
                patch_output, teacher_temp=0.04, n_masked_patches_tensor=torch.tensor([16])
            )
            self.assertTrue(torch.equal(patch_output[0], patch_output_clone[0]))
            self.assertTrue(torch.equal(patch_output[1], patch_output_clone[1]))

    def test_i_jepa(self) -> None:
        batch_size = 4
        size = (192, 192)
        backbone = registry.net_factory("vit_s16", 0, size=size)
        input_size = (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride)
        predictor = i_jepa.VisionTransformerPredictor(
            input_size,
            backbone.embedding_size,
            128,
            512,
            8,
            2,
            0.0,
        )
        mask_generator = i_jepa.MultiBlockMasking(
            input_size,
            enc_mask_scale=(0.85, 1.0),
            pred_mask_scale=(0.15, 0.25),
            aspect_ratio=(0.75, 1.5),
            n_enc=2,
            n_pred=3,
            min_keep=10,
            allow_overlap=False,
        )
        masks = mask_generator(batch_size)
        enc_masks = masks[0]
        pred_masks = masks[1]

        x = torch.rand(batch_size, DEFAULT_NUM_CHANNELS, *size)

        #
        # Simulate a full step
        #

        # Target encoder
        h = backbone.masked_encoding_omission(x)["tokens"]
        h = h[:, backbone.num_special_tokens :, :]  # Remove special tokens
        h = i_jepa.apply_masks(h, pred_masks)
        h = i_jepa.repeat_interleave_batch(h, batch_size, repeat=len(enc_masks))

        # Context
        z = torch.concat([backbone.masked_encoding_omission(x, enc_mask)["tokens"] for enc_mask in enc_masks], dim=0)
        z = z[:, backbone.num_special_tokens :, :]  # Remove special tokens
        z = predictor(z, enc_masks, pred_masks)

        loss = F.smooth_l1_loss(z, h)

        self.assertEqual(h.size(0), batch_size * len(enc_masks) * len(pred_masks))
        self.assertEqual(z.size(0), batch_size * len(enc_masks) * len(pred_masks))

        self.assertEqual(h.size(2), backbone.embedding_size)
        self.assertEqual(z.size(2), backbone.embedding_size)

        self.assertEqual(h.size(1), z.size(1))

        self.assertEqual(loss.ndim, 0)

    def test_ibot(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("vit_b32", 0)
        backbone.set_dynamic_size()
        net = ibot.iBOT(
            backbone,
            config={
                "out_dim": 128,
                "norm_last_layer": True,
                "num_layers": 3,
                "hidden_dim": 512,
                "bottleneck_dim": 256,
                "patch_out_dim": 192,
                "shared_head": False,
            },
        )

        # Test network
        images = [
            # Global
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
            # Local
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 96, 96)),
        ]

        mask_generator = masking.BlockMasking(
            (128 // backbone.stem_stride, 128 // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)

        embedding_g, features_g = net(torch.concat(images[:2], dim=0), masks=masks)
        self.assertFalse(torch.isnan(embedding_g).any())
        self.assertFalse(torch.isnan(features_g).any())
        self.assertEqual(features_g.size(), (batch_size * 2, (128 // 32) ** 2, 192))
        self.assertEqual(embedding_g.size(), (batch_size * 2, 128))

        embedding_l, features_l = net(torch.concat(images[2:], dim=0), masks=None)
        self.assertFalse(torch.isnan(embedding_l).any())
        self.assertFalse(torch.isnan(features_l).any())
        self.assertEqual(features_l.size(), (batch_size * 4, (96 // 32) ** 2, 192))
        self.assertEqual(embedding_l.size(), (batch_size * 4, 128))

        ibot_loss = ibot.iBOTLoss(
            128,
            192,
            num_global_crops=2,
            num_local_crops=4,
            warmup_teacher_temp=0.1,
            teacher_temp=0.9,
            warmup_teacher_temp2=0.2,
            teacher_temp2=0.99,
            warmup_teacher_temp_epochs=5,
            epochs=100,
            student_temp=0.5,
            center_momentum=0.98,
            center_momentum2=0.97,
            lambda1=0.2,
            lambda2=0.1,
            mim_start_epoch=1,
        )

        loss = ibot_loss.forward(
            embedding_g,
            features_g,
            torch.rand_like(embedding_g),
            torch.rand_like(features_g),
            student_local_embedding=embedding_l,
            student_mask=masks,
            epoch=2,
        )

        self.assertFalse(torch.isnan(loss["all"]).any())
        self.assertFalse(torch.isnan(loss["embedding"]).any())
        self.assertFalse(torch.isnan(loss["features"]).any())

        self.assertEqual(loss["all"].ndim, 0)
        self.assertEqual(loss["embedding"].ndim, 0)
        self.assertEqual(loss["features"].ndim, 0)

    def test_mmcr(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v1_50", 0)
        net = mmcr.MMCR(backbone, config={"projector_dims": [512, 512, 512]})
        mmcr_loss = mmcr.MMCRMomentumLoss(0.0, 2)

        # Test network
        out, out_m = net(torch.rand((batch_size, 2, DEFAULT_NUM_CHANNELS, 128, 128)))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 2)
        self.assertFalse(torch.isnan(out_m).any())
        self.assertEqual(out_m.ndim, 2)

        # Test loss
        loss = mmcr_loss(out, out_m)
        self.assertFalse(torch.isnan(loss).any())
        self.assertEqual(loss.ndim, 0)

    def test_simclr(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v2_18", 0)
        net = simclr.SimCLR(
            backbone,
            config={
                "projection_dim": 96,
                "projection_hidden_dim": 128,
                "temperature": 0.1,
            },
        )

        # Test network
        out = net(
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
        )
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_sscd(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v2_18", 512)
        net = sscd.SSCD(backbone)
        out = net(torch.rand(batch_size, DEFAULT_NUM_CHANNELS, 128, 128))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 2)

    def test_vicreg(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v1_18", 0)
        net = vicreg.VICReg(
            backbone,
            config={
                "mlp_dim": 128,
                "batch_size": batch_size,
                "sim_coeff": 0.1,
                "std_coeff": 0.1,
                "cov_coeff": 0.1,
            },
        )

        # Test network
        out = net(
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
            torch.rand((batch_size, DEFAULT_NUM_CHANNELS, 128, 128)),
        )
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)



"""
Fast Gradient Sign Method (FGSM)

Paper "Explaining and Harnessing Adversarial Examples", https://arxiv.org/abs/1412.6572
"""

from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.adversarial.base import AttackResult
from birder.adversarial.base import attack_success
from birder.adversarial.base import clamp_normalized
from birder.adversarial.base import pixel_eps_to_normalized
from birder.adversarial.base import predict_labels
from birder.adversarial.base import validate_target
from birder.data.transforms.classification import RGBType


class FGSM:
    def __init__(self, net: nn.Module, eps: float, *, rgb_stats: RGBType) -> None:
        self.net = net.eval()
        self.eps = eps
        self.rgb_stats = rgb_stats

    def __call__(self, input_tensor: torch.Tensor, target: Optional[torch.Tensor]) -> AttackResult:
        inputs = input_tensor.detach().clone()
        inputs.requires_grad_(True)

        logits = self.net(inputs)
        targeted = target is not None
        if targeted is True:
            target = validate_target(target, inputs.shape[0], logits.shape[1], inputs.device)
        else:
            target = predict_labels(logits)

        loss = F.cross_entropy(logits, target)
        (grad,) = torch.autograd.grad(loss, inputs, retain_graph=False, create_graph=False)
        eps_norm = pixel_eps_to_normalized(self.eps, self.rgb_stats, device=inputs.device, dtype=inputs.dtype)

        # Targeted steps descend toward target, untargeted ascend away from original
        if targeted is True:
            direction = -1.0
        else:
            direction = 1.0

        perturbation = direction * eps_norm * grad.sign()
        adv_inputs = clamp_normalized(inputs + perturbation, self.rgb_stats)
        with torch.no_grad():
            adv_logits = self.net(adv_inputs)

        success = attack_success(
            logits.detach(),
            adv_logits,
            targeted,
            target=target if targeted else None,
        )

        return AttackResult(
            adv_inputs=adv_inputs,
            adv_logits=adv_logits,
            perturbation=adv_inputs - inputs,
            logits=logits.detach(),
            success=success,
        )



"""
MLP probe for multi-label classification evaluation

Trains a 2-layer MLP on pre-extracted embeddings to predict binary traits.
Architecture follows the AwA2 (Animals with Attributes 2) design:
Linear -> Dropout -> Linear (no activation between layers).
"""

import logging

import numpy as np
import numpy.typing as npt
import torch
from sklearn.metrics import f1_score
from torch import nn
from torch.utils.data import DataLoader
from torch.utils.data import TensorDataset

logger = logging.getLogger(__name__)


class MLPProbe(nn.Module):
    def __init__(self, input_dim: int, num_classes: int, hidden_dim: int = 512, dropout: float = 0.5) -> None:
        super().__init__()
        self.model = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, num_classes),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


# pylint: disable=too-many-locals
def train_mlp(
    train_features: npt.NDArray[np.float32],
    train_labels: npt.NDArray[np.float32],
    num_classes: int,
    device: torch.device,
    epochs: int,
    batch_size: int,
    lr: float = 1e-4,
    step_size: int = 20,
    hidden_dim: int = 512,
    dropout: float = 0.5,
    seed: int = 0,
) -> MLPProbe:
    """
    Train MLP probe on embeddings for multi-label classification

    Parameters
    ----------
    train_features
        Training features of shape (n_train, embedding_dim).
    train_labels
        Training labels of shape (n_train, num_classes), float values 0/1.
    num_classes
        Number of output classes (traits).
    device
        Device to train on.
    epochs
        Number of training epochs.
    batch_size
        Batch size for training.
    lr
        Learning rate.
    step_size
        Number of epochs between learning rate decay steps.
    hidden_dim
        Hidden layer dimension.
    dropout
        Dropout probability.
    seed
        Random seed for reproducibility.

    Returns
    -------
    Trained MLPProbe model.
    """

    torch.manual_seed(seed)
    np.random.seed(seed)

    input_dim = train_features.shape[1]
    model = MLPProbe(input_dim, num_classes, hidden_dim, dropout).to(device)

    x_train = torch.from_numpy(train_features).float()
    y_train = torch.from_numpy(train_labels).float()

    dataset = TensorDataset(x_train, y_train)
    loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

    criterion = nn.BCEWithLogitsLoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=0.5)

    model.train()
    for epoch in range(epochs):
        total_loss = 0.0
        for batch_x, batch_y in loader:
            batch_x = batch_x.to(device)
            batch_y = batch_y.to(device)

            optimizer.zero_grad()
            logits = model(batch_x)
            loss = criterion(logits, batch_y)
            loss.backward()
            optimizer.step()

            total_loss += loss.item() * batch_x.size(0)

        scheduler.step()

        if (epoch + 1) % 10 == 0:
            avg_loss = total_loss / len(dataset)
            logger.debug(f"Epoch {epoch + 1}/{epochs} - Loss: {avg_loss:.4f}")

    return model  # type: ignore[no-any-return]


def evaluate_mlp(
    model: MLPProbe,
    test_features: npt.NDArray[np.float32],
    test_labels: npt.NDArray[np.float32],
    batch_size: int = 128,
    device: torch.device = torch.device("cpu"),
) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.int_], float]:
    """
    Evaluate MLP probe on test set

    Parameters
    ----------
    model
        Trained MLPProbe model.
    test_features
        Test features of shape (n_test, embedding_dim).
    test_labels
        Test labels of shape (n_test, num_classes), float values 0/1.
    batch_size
        Batch size for inference.
    device
        Device to run on.

    Returns
    -------
    y_pred
        Predicted labels of shape (n_test, num_classes), int values 0/1.
    y_true
        True labels of shape (n_test, num_classes), int values 0/1.
    macro_f1
        Macro-averaged F1 score across all classes.
    """

    x_test = torch.from_numpy(test_features).float()
    y_test = torch.from_numpy(test_labels).float()

    dataset = TensorDataset(x_test, y_test)
    loader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

    model.to(device)
    model.eval()
    all_preds: list[torch.Tensor] = []
    all_labels: list[torch.Tensor] = []

    with torch.inference_mode():
        for batch_x, batch_y in loader:
            batch_x = batch_x.to(device)
            logits = model(batch_x)
            preds = (torch.sigmoid(logits) > 0.5).int()
            all_preds.append(preds.cpu())
            all_labels.append(batch_y.int())

    y_pred = torch.concat(all_preds, dim=0).numpy()
    y_true = torch.concat(all_labels, dim=0).numpy()
    macro_f1 = f1_score(y_true, y_pred, average="macro", zero_division=0.0)

    return (y_pred, y_true, float(macro_f1))



import argparse
import fnmatch
import logging
from typing import Any
from typing import Optional

import numpy as np
import numpy.typing as npt
import polars as pl
import polars.datatypes.classes
from rich.console import Console
from rich.table import Table

from birder.common import cli
from birder.conf import settings
from birder.results.detection import Results
from birder.results.gui import ConfusionMatrix
from birder.tools.results import print_most_confused_pairs

logger = logging.getLogger(__name__)


def compare_results(results_dict: dict[str, Results]) -> pl.DataFrame:
    result_list = []
    for name, results in results_dict.items():
        report_df = results.detailed_report()
        total_objects = int(report_df["Objects"].sum()) if report_df.is_empty() is False else 0
        result_entry = {
            "File name": name,
            "mAP": results.map,
            "mAP_50": results.metrics_dict["map_50"],
            "mAP_75": results.metrics_dict["map_75"],
            "mAP_small": results.metrics_dict["map_small"],
            "mAP_medium": results.metrics_dict["map_medium"],
            "mAP_large": results.metrics_dict["map_large"],
            "mAR_1": results.metrics_dict["mar_1"],
            "mAR_10": results.metrics_dict["mar_10"],
            "mAR_100": results.metrics_dict["mar_100"],
            "mAR_small": results.metrics_dict["mar_small"],
            "mAR_medium": results.metrics_dict["mar_medium"],
            "mAR_large": results.metrics_dict["mar_large"],
            "Samples": len(results),
            "Objects": total_objects,
        }
        result_list.append(result_entry)

    return pl.DataFrame(result_list)


def print_per_class_report(results_dict: dict[str, Results], classes: list[str]) -> None:
    console = Console()

    all_classes = []
    for results in results_dict.values():
        for cls in classes:
            all_classes.extend(fnmatch.filter(results.label_names, cls))

    classes = sorted(list(set(all_classes)))

    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("File name")
    table.add_column("Class name", style="dim")
    table.add_column("mAP", justify="right")
    table.add_column("mAR 100", justify="right")
    table.add_column("Objects", justify="right")

    per_class_mar = {
        name: dict(zip(results.metrics_dict["classes"], results.metrics_dict["mar_100_per_class"]))
        for name, results in results_dict.items()
    }

    for cls in classes:
        for name, results in results_dict.items():
            report_df = results.detailed_report()
            row = report_df.filter(pl.col("Class name") == cls)
            if row.is_empty() is True:
                continue

            class_num = int(row["Class"][0])
            mar_value = per_class_mar[name].get(class_num)
            mar_msg = "n/a" if mar_value is None else f"{mar_value:.4f}"

            map_msg = f"{row['mAP'][0]:.4f}"
            if row["mAP"][0] < 0.4:
                map_msg = "[red1]" + map_msg + "[/red1]"
            elif row["mAP"][0] < 0.5:
                map_msg = "[dark_orange]" + map_msg + "[/dark_orange]"

            table.add_row(
                name,
                row["Class name"][0],
                map_msg,
                mar_msg,
                f"{row['Objects'][0]}",
            )

    console.print(table)


def print_report(results_dict: dict[str, Results]) -> None:
    if len(results_dict) == 1:
        results = next(iter(results_dict.values()))
        results.pretty_print()
        return

    results_df = compare_results(results_dict)
    console = Console()
    table = Table(show_header=True, header_style="bold dark_magenta")
    for idx, column in enumerate(results_df.columns):
        if idx == 0:
            table.add_column(column)
        else:
            table.add_column(column, justify="right")

        if isinstance(results_df[column].dtype, polars.datatypes.classes.FloatType):
            results_df = results_df.with_columns(
                pl.col(column).map_elements(lambda x: f"{x:.4f}", return_dtype=pl.String)
            )
        else:
            results_df = results_df.with_columns(pl.col(column).cast(pl.String))

    for row in results_df.iter_rows():
        table.add_row(*row)

    console.print(table)
    console.print("\n")


def confusion_matrix_data(
    results: Results,
    score_threshold: float,
    iou_threshold: float,
    classes: list[str],
    errors_only: bool,
) -> tuple[npt.NDArray[np.int_], list[str]]:
    cnf_matrix = results.confusion_matrix(score_threshold=score_threshold, iou_threshold=iou_threshold)
    label_names = results.label_names

    selected_indices: Optional[list[int]] = None
    if len(classes) > 0:
        selected_indices = []
        for pattern in classes:
            selected_indices.extend([idx for idx, name in enumerate(label_names) if fnmatch.fnmatch(name, pattern)])

        selected_indices = np.unique(selected_indices).tolist()

    elif errors_only is True:
        row_sums = cnf_matrix.sum(axis=1)
        col_sums = cnf_matrix.sum(axis=0)
        diag = np.diag(cnf_matrix)
        selected_indices = np.where((row_sums != diag) | (col_sums != diag))[0].tolist()
        if len(selected_indices) == 0:
            logger.warning("No mistakes found, showing full confusion matrix")
            return (cnf_matrix, label_names)

    if selected_indices is None:
        return (cnf_matrix, label_names)

    if 0 not in selected_indices:
        selected_indices.insert(0, 0)

    selected_indices = sorted(list(set(selected_indices)))

    cnf_matrix = cnf_matrix[np.ix_(selected_indices, selected_indices)]
    label_names = [label_names[idx] for idx in selected_indices]

    return (cnf_matrix, label_names)


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "det-results",
        allow_abbrev=False,
        help="read and process detection result files",
        description="read and process detection result files",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools det-results "
            "results/faster_rcnn_coco_csp_resnet_50_imagenet1k_91_e0_640px_5000.json --print\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--print", default=False, action="store_true", help="print results table")
    subparser.add_argument("--short-print", default=False, action="store_true", help="print results")
    subparser.add_argument("--save-summary", default=False, action="store_true", help="save results summary as csv")
    subparser.add_argument("--summary-suffix", type=str, help="add suffix to summary file")
    subparser.add_argument("--classes", default=[], type=str, nargs="+", help="class names to compare")
    subparser.add_argument("--cnf", default=False, action="store_true", help="plot confusion matrix")
    subparser.add_argument(
        "--cnf-errors-only",
        default=False,
        action="store_true",
        help="show only classes with mistakes at the confusion matrix",
    )
    subparser.add_argument("--cnf-save", default=False, action="store_true", help="save confusion matrix as csv")
    subparser.add_argument(
        "--cnf-score-threshold", type=float, default=0.5, help="detection score threshold for confusion matrix"
    )
    subparser.add_argument(
        "--cnf-iou-threshold", type=float, default=0.5, help="iou threshold for confusion matrix matching"
    )
    subparser.add_argument("--most-confused", default=False, action="store_true", help="print most confused pairs")
    subparser.add_argument("result_files", type=str, nargs="+", help="result files to process")
    subparser.set_defaults(func=main)


# pylint: disable=too-many-branches
def main(args: argparse.Namespace) -> None:
    results_dict: dict[str, Results] = {}
    for results_file in args.result_files:
        results = Results.load(results_file)
        result_name = results_file.split("/")[-1]
        results_dict[result_name] = results

    if args.print is True:
        print_report(results_dict)
        if len(args.classes) > 0:
            print_per_class_report(results_dict, args.classes)

    if args.short_print is True:
        for name, results in results_dict.items():
            print(f"{name}: {results}\n")

    if args.save_summary is True:
        if args.summary_suffix is not None:
            summary_path = settings.RESULTS_DIR.joinpath(f"summary_{args.summary_suffix}.csv")
        else:
            summary_path = settings.RESULTS_DIR.joinpath("summary.csv")

        if summary_path.exists() is True:
            logger.warning(f"Summary already exists '{summary_path}', skipping...")
        else:
            logger.info(f"Writing results summary at '{summary_path}...")
            results_df = compare_results(results_dict)
            results_df.write_csv(summary_path)

    if args.cnf is True:
        if len(results_dict) > 1:
            logger.warning("Cannot compare confusion matrix, processing only the first file")

        results = next(iter(results_dict.values()))
        cnf_matrix, label_names = confusion_matrix_data(
            results, args.cnf_score_threshold, args.cnf_iou_threshold, args.classes, args.cnf_errors_only
        )
        title = f"Confusion matrix (score >= {args.cnf_score_threshold:.2f}, IoU >= {args.cnf_iou_threshold:.2f})"
        ConfusionMatrix(cnf_matrix, label_names, title=title).show()

    if args.cnf_save is True:
        for results_file, results in results_dict.items():
            filename = f"{results_file.rsplit('.', 1)[0]}_confusion_matrix.csv"
            cnf_matrix = results.confusion_matrix(
                score_threshold=args.cnf_score_threshold, iou_threshold=args.cnf_iou_threshold
            )
            ConfusionMatrix(cnf_matrix, results.label_names).save(filename)

    if args.most_confused is True:
        if len(results_dict) > 1:
            logger.warning("Cannot compare, processing only the first file")

        results = next(iter(results_dict.values()))
        most_confused_df = results.most_confused_pairs(
            n=10, score_threshold=args.cnf_score_threshold, iou_threshold=args.cnf_iou_threshold
        )
        print_most_confused_pairs(most_confused_df)
